#include <iostream.h>

#include "Game.H"

#include "Common.H"
#include "AIObject.H"
#include "AIState.H"
#include "AIGUI.H"

AIObject::AIObject(GameObj* gob) {
    this->gob = gob;
    id = state.last_aiob_id++;
    dead = false;
    owner = *gob->sod.owner;
    ug = new Unit_gui(this);
}

void AIObject::refresh() {
    pos.x = *gob->sod.x;
    pos.y = *gob->sod.y;
}

void AIObject::redraw(AIGUI* gui) {
    gui->line(pos.x-3, pos.y-3, pos.x+3, pos.y+3, ClWhite);
    gui->line(pos.x-3, pos.y+3, pos.x+3, pos.y-3, ClWhite);
}

bool AIObject::is_moving() {
    return *gob->sod.speed > 0;
}

float AIObject::dist_to(AIObject* aiob) {
    return gob->distance_to(*aiob->gob);
}

Asheep::Asheep(GameObj* gob) : AIObject(gob) {}

void Asheep::redraw(AIGUI* gui) {
    gui->circle(pos, state.SHEEP_SIZE, ClGray, true);
}

AMarine::AMarine(GameObj* gob) : AIObject(gob) {
    range = gob->component("weapon")->get_int("max_ground_range");
}

void AMarine::refresh() {
    AIObject::refresh();
    prev_hp = hp;
    hp = gob->get_int("hp");
}

void AMarine::redraw(AIGUI* gui) {
    ug->show_dmg(gui, ((80 - hp) * S_ARR_MAX) / 80);
    AMarine* e = nearby_enemy();
    if (e)
	if ((e->id + state.gameTime) % state.MARINE_COOLDOWN > 0) // oszustwo!
	    e = NULL;
    gui->circle(pos, state.MARINE_SIZE, e ? ClRed : ClMaroon, true);
}

AMarine* AMarine::nearby_enemy() {
    AMarine* e = NULL;
    int min_hp = state.MARINE_MAX_HP + 1;
    UnitGroup* g = is_ally() ? state.oppMarines : state.myMarines;
    FORALL(g->v, it) {
	AMarine* m = (AMarine*)*it;
	if (m->dead) // nie powinno sie zdarzyc
	    continue;
	if (dist_to(m) < range)
	    if (m->hp < min_hp) {
		min_hp = m->hp;
		e = m;
	    }
    }
    return e;
}

MyMarine::MyMarine(GameObj* gob) : AMarine(gob) {
    stop();
}

void MyMarine::refresh() {
    AMarine::refresh();
    cooldown = state.gameTime - gob->component("weapon")->get_int("last_fired");
    if (direct_target)
	if (direct_target->dead)
	    direct_target = NULL;
}

void MyMarine::redraw(AIGUI* gui) {
    if (L2_DEBUG && (dest.x >= 0)) gui->line(pos, dest, ClPink);
    AMarine::redraw(gui);
    gui->circle(pos, state.MARINE_SIZE, (cooldown != state.MARINE_COOLDOWN-1) ? ClNavy : ClBlue, true);
    if (AI_DEBUG) {
        if (idle) gui->line(pos, TPoint(pos.x, pos.y-5), ClWhite);
	if (target) gui->line(pos, TPoint(pos.x+5, pos.y-5), ClGray);
        if (in_move) gui->line(pos, TPoint(pos.x+5, pos.y), ClGray);
	if (is_moving()) gui->line(pos, TPoint(pos.x+5, pos.y+5), ClGray);
        if (around) gui->line(pos, TPoint(pos.x-5, pos.y+5), ClBlue);
	if (direct_target) gui->line(pos, TPoint(pos.x+5, pos.y-5), ClGray);
        if (go_attack) gui->line(pos, TPoint(pos.x-5, pos.y), ClGray);
	if (bumps) gui->line(pos, TPoint(pos.x, pos.y+5), ClGray);
    }
    if ((cooldown <= state.MARINE_COOLDOWN) && (direct_target)) {
	float p1 = (float)cooldown / (float)(state.MARINE_COOLDOWN + 1);
	float p2 = (float)(cooldown + 1) / (float)(state.MARINE_COOLDOWN + 1);
	float x1 = ((float)(pos.x) * (1 - p1) + (float)(direct_target->pos.x) * p1);
	float y1 = ((float)(pos.y) * (1 - p1) + (float)(direct_target->pos.y) * p1);
	float x2 = ((float)(pos.x) * (1 - p2) + (float)(direct_target->pos.x) * p2);
	float y2 = ((float)(pos.y) * (1 - p2) + (float)(direct_target->pos.y) * p2);
	gui->line((int)x1, (int)y1, (int)x2, (int)y2, ClNavy);
    }
}

void MyMarine::low_stop() {
    Vector<sint4> params;
    gob->set_action("stop", params);
//    cout << id << ": low_stop\n";
}

void MyMarine::low_move(TPoint pos) {
    Vector<sint4> params;
    params.push_back(pos.x);
    params.push_back(pos.y);
    params.push_back(*gob->sod.max_speed);
    gob->set_action("move", params);
//    cout << id <<  ": low_move\n";
}

void MyMarine::low_attack(AMarine* aiob) {
    if (dist_to(aiob) > range) {
	cerr << "BLAD: cel poza zasiegiem!\n";
	return;
    }
    Vector<sint4> params;
    params.push_back(state.gsm->get_game().get_cplayer_info().get_id(aiob->gob));
//    cout << id << ": low_attack" << endl;
    gob->component("weapon")->set_action("attack", params);
    direct_target = aiob;
}

void MyMarine::update() {
    TPoint tmp_dest = TPoint(-1, -1);
    AMarine* tmp_target = NULL;
    if (direct_target)
	if (direct_target->dead)
	    direct_target = NULL;
    if (in_move) {
	if (target && !around) {
	    if (target->dead) {
		target = NULL;
		cout << id << " stopped: target died" << endl;
		stop();
		return;
	    }
	    if (is_moving()) {
		dest = target->pos;
		tmp_dest = dest;
	    }
	}
	if (!is_moving() && !direct_target) {
	    if ((dist(pos, dest) < 3) && !target) {
		cout << id << " stopped: reached dest." << endl;
		stop();
		return;
	    }
	    if (dist(pos, dest) == prev_dist) {
		bumps++;
		tot_bumps++;
		if ((bumps > 20) || (tot_bumps > max_bumps)) {
		    cout << id << " stopped: too many colisions" << endl;
		    stop();
		    return;
		} else {
		    int x1 = pos.x + rand() % 15 - 7;
		    int y1 = pos.y + rand() % 15 - 7;
		    around = true;
		    tmp_dest = TPoint(x1, y1);
		}
	    } else {
		prev_dist = dist(pos, dest);
		bumps = 0;//?
		around = false;
		tmp_dest = dest;
	    }
	}
	if (go_attack && target)
	    if (dist_to(target) < range)
		tmp_target = (AMarine*)target;
    }
    if (!tmp_target && agressive)
	tmp_target = nearby_enemy();
    if (tmp_target) {
	if (is_moving()) { //?
	    low_stop();
	} else {
	    low_attack(tmp_target);
	}
    } else {
	if (tmp_dest.x >= 0)
	    low_move(tmp_dest);
    }
}

void MyMarine::stop() {
    direct_target = NULL;
    prev_dist = -1;
    bumps = 0;
    tot_bumps = 0;
    max_bumps = 100; // default
    around = false;
    in_move = false;
    go_attack = false;
    agressive = false;
    idle = true;
    dest = TPoint(-1, -1);
    target = NULL;
    low_stop();
}

void MyMarine::moveTo(TPoint p) {
    stop();
    idle = false;
    in_move = true;
    dest = p;
    prev_dist = dist(pos, dest);
    low_move(dest);
}

void MyMarine::follow(AIObject* aiob) {
    moveTo(aiob->pos);
    target = aiob;
}

void MyMarine::attack(AMarine* aiob) {
    if (!aiob->is_enemy())
	ERR("Cannot attack allies nor neutrals");
    if (aiob->dead)
	ERR("Cannot attack dead unit");
    follow(aiob);
    go_attack = true;
}

void MyMarine::holdPosition() {
    stop();
    agressive = true;
}

void MyMarine::autoAttack(TPoint p) {
    moveTo(p);
    agressive = true;
}

void MyMarine::autoAttack(AMarine* aiob) {
    follow(aiob);
    agressive = true;
}
