#include <iostream.h>

#include "GUI.H"
#include "GameObj.H"

#include "AIGUI.H"
#include "AIState.H"

void AIGUI::add_options() {
    GUI::add_options();
}

AIGUI::AIGUI() {
    cout << "Initializing GUI...\n";
    GUI::GUI();
    minimap.h = 10;
    init(WWidth, WHeight, state.gsm->get_game());
    adjust_view_mag(-2);
    gv2 = views[0];
}

void AIGUI::draw_digth(const GfxView2& gv, int x, int y, int w, int h, int n, const RGB& c) {
    if ((n < 0) || (n > 9))
	ERR("zla wartosc cyfry.");
    int dtab[10][7] = {
	{1, 1, 1, 1, 1, 1, 0},
	{0, 1, 1, 0, 0, 0, 0},
	{1, 1, 0, 1, 1, 0, 1},
	{1, 1, 1, 1, 0, 0, 1},
	{0, 1, 1, 0, 0, 1, 1},
	{1, 0, 1, 1, 0, 1, 1},
	{1, 0, 1, 1, 1, 1, 1},
	{1, 1, 1, 0, 0, 0, 0},
	{1, 1, 1, 1, 1, 1, 1},
	{1, 1, 1, 1, 0, 1, 1}
    };
    int dpos[7][4] = {
	{0, 0, 1, 0},
	{1, 0, 1, 1},
	{1, 1, 1, 2},
	{0, 2, 1, 2},
	{0, 1, 0, 2},
	{0, 0, 0, 1},
	{0, 1, 1, 1}
    };
    h /= 2;
    for (int i=0; i<7; i++)
	if (dtab[n][i])
	    draw_line(gv, x+dpos[i][0]*w, y+dpos[i][1]*h, x+dpos[i][2]*w, y+dpos[i][3]*h, c);
}

void AIGUI::draw_int(int x, int y, int w0, int h, int n, const RGB& c) {
    do {
	draw_digth(gv2, x, y, w0 * 4 / 5, h, n % 10, c);
	n = n / 10;
	x -= w0;
    } while (n > 0);
}

void AIGUI::line(int x1, int y1, int x2, int y2, const RGB& c) {
    draw_line(gv2, x1, y1, x2, y2, c);
}

void AIGUI::circle(int x, int y, int r, const RGB& c, bool filled) {
    draw_circle(gv2, x, y, r, c, filled);
}

void AIGUI::line(TPoint p1, TPoint p2, const RGB& c) {
    draw_line(gv2, p1.x, p1.y, p2.x, p2.y, c);
}

void AIGUI::circle(TPoint s, int r, const RGB& c, bool filled) {
    draw_circle(gv2, s.x, s.y, r, c, filled);
}

void AIGUI::rect(int x, int y, int w, int h, const RGB& fc) {
    line(x, y, x+w, y, fc);
    line(x, y, x, y+h, fc);
    line(x+w, y, x+w, y+h, fc);
    line(x, y+h, x+w, y+h, fc);
}

void AIGUI::filledRect(int x, int y, int w, int h, const RGB& bc) {
    for (int i=y+1; i<y+h; i++)
	line(x, i, x+w, i, bc);
}

void AIGUI::display(const GameObj* gob1, const GameObj* gob2) {
        if (SDL_MUSTLOCK(screen) && (SDL_LockSurface(screen) < 0)) {
	cerr << "\"Can't lock screen: " << SDL_GetError() << "\"\n";
	return;
    }
    gv2 = views[0];
    FORALL(views, it) {
	correct_view();
	GUI::display(*it, gob1, gob2);
    }
    filledRect(0, 0, state.MAP_WIDTH, state.MAP_HEIGHT, ClBackground);
    rect(0, 0, state.MAP_WIDTH, state.MAP_HEIGHT, ClWhite);
    FORALL(state.allUnits->v, it)
	(*it)->draw_bg(this);
    FORALL(state.allUnits->v, it)
	(*it)->redraw(this);
    if (state.gameTime % 2) {
	if (gob1)
    	    circle(*gob1->sod.x, *gob1->sod.y, 5, ClWhite);
	if (gob2)
	    circle(*gob2->sod.x, *gob2->sod.y, 5, ClWhite);
    }
    draw_gamestat();
    if (SDL_MUSTLOCK(screen))
	SDL_UnlockSurface(screen);
    SDL_UpdateRect(screen, 0, 0, 0, 0);
}

void AIGUI:: draw_gamestat() {
    draw_int(50, 20, 25, 50, state.myMarines->size, ClBlue);
    draw_int(980, 20, 25, 50, state.oppMarines->size, ClRed);
    if (SHOW_TOTAL_LIFE) {
        int d1 = 0;
	int d2 = 0;
	FORALL(state.myMarines->v, i1) {
	    bool d2p = false;
	    FORALL(state.oppMarines->v, i2)
		if (((AMarine*)(*i1))->dist_to((AMarine*)(*i2)) < ((AMarine*)(*i1))->range)
		    d2p = true;
	    if (d2p) d2++;
	}
	FORALL(state.oppMarines->v, i1) {
	    bool d1p = false;
	    FORALL(state.myMarines->v, i2)
		if (((AMarine*)(*i1))->dist_to((AMarine*)(*i2)) < ((AMarine*)(*i1))->range)
		    d1p = true;
	    if (d1p) d1++;
	}
	filledRect(100, 20, state.myMarines->life() / 5, 10, ClNavy);
	filledRect(100, 32, state.oppMarines->life() / 5, 10, ClMaroon);
	if (state.myMarines->life() / 5 > d1)
	    filledRect(100+state.myMarines->life() / 5 - d1, 20, d1, 10, ClBlue);
	if (state.oppMarines->life() / 5 > d2)
	    filledRect(100+state.oppMarines->life() / 5 - d2, 32, d2, 10, ClRed);
	rect(100, 20, 800, 10, ClGray);
	rect(100, 32, 800, 10, ClGray);
    }
    if (state.behind)
	circle(TPoint(state.MAP_WIDTH-50, state.MAP_HEIGHT-50), 30, ClMaroon, true);
//    draw_int(20, state.MAP_HEIGHT - 40, 10, 20, state.AIstage, ClBlack);
}

Unit_gui::Unit_gui(AIObject* a) {
    aiob = a;
    for (int i=0; i<S_ARR_MAX; i++)
	s_arr[i][0] = -1;
    heat = 0;
}

int rand30(int n) {
    int r = n - rand() % 30;
    return (r < 0) ? 0 : r;
}

void Unit_gui::show_dmg(AIGUI* gui, int n) {
    int m = S_ARR_MAX-n-1;
    if (m < 0) m = 0;
    s_arr[m][0] = aiob->pos.x;
    s_arr[m][1] = aiob->pos.y;
    s_arr[m][2] = m + rand() % m;
    for (int k=0; k<4; k++)
	for (int i=S_ARR_MAX-1; i>m; i--)
	    for (int j=0; j<3; j++)
		s_arr[i][j] = s_arr[i-1][j];
    for (int i=0; i<S_ARR_MAX; i++)
	if (s_arr[i][0] >= 0) {
	    s_arr[i][0] += rand() % 5 - 2;
	    if (rand() % (s_arr[i][2] + 1) < 25)
		s_arr[i][1] -= rand() % 4 - 1;
	    s_arr[i][2] += rand() % 3;
	}
    for (int i=0; i<S_ARR_MAX; i++)
	if (s_arr[i][0] >= 0) {
	    gui->circle(TPoint(s_arr[i][0], s_arr[i][1]), 1,
		RGB(rand30(s_arr[i][2]), rand30(s_arr[i][2]),
		rand30(s_arr[i][2])), true);
	}
    s_arr[0][0] = -1;
}
