#include <iostream.h>
#include <math.h>

#include "Game.H"
#include "GUI.H"

#include "Common.H"
#include "AIObject.H"
#include "AIState.H"
#include "AIGUI.H"

AIObject::AIObject(GameObj* gob) {
    this->gob = gob;
    id = state.last_aiob_id++;
    dead = false;
    owner = *gob->sod.owner;
    ug = new Unit_gui(this);
    for (int i=0; i<PPOS_LEN; i++)
	prev_pos[i] = nowhere;
}

void AIObject::refresh() {
    if (is_moving()) {
	for (int i=PPOS_LEN-1; i>0; i--)
	    prev_pos[i] = prev_pos[i-1];
	prev_pos[0] = pos;
    }
    pos.x = *gob->sod.x;
    pos.y = *gob->sod.y;
}

void AIObject::redraw(AIGUI* gui) {
    gui->line(pos.x-3, pos.y-3, pos.x+3, pos.y+3, ClWhite);
    gui->line(pos.x-3, pos.y+3, pos.x+3, pos.y-3, ClWhite);
}

bool AIObject::is_moving() {
    return *gob->sod.speed > 0;
}

float AIObject::dist_to(AIObject* aiob) {
    return gob->distance_to(*aiob->gob);
}

TPoint AIObject::next_pos(int ticks) {
    if (prev_pos[PPOS_LEN] == nowhere)
	return pos;
    return pos + (pos - prev_pos[PPOS_LEN-1]) * (ticks / PPOS_LEN);
}

TPoint halfWay(TPoint pos1, TPoint pos2, int d) {
    if (pos1 == pos2)
	return pos2;
    return pos1 + (pos2 - pos1) * ((float)d / (float)(dist_f(pos1, pos2)));
}

Asheep::Asheep(GameObj* gob) : AIObject(gob) {}

void Asheep::redraw(AIGUI* gui) {
    gui->circle(pos, state.SHEEP_SIZE, ClGray, true);
}

AMarine::AMarine(GameObj* gob) : AIObject(gob) {
    range = gob->component("weapon")->get_int("max_ground_range");
    cooldown = 0;
    direct_target = NULL;
}

void AMarine::refresh() {
    AIObject::refresh();
    prev_hp = hp;
    hp = gob->get_int("hp");
}

void draw_bullet(AIGUI* gui, AMarine* m, const RGB& col) {
    if ((m->direct_target == NULL) || (m->cooldown > state.MARINE_COOLDOWN))
	return;
    float p1 = (float)(m->cooldown) / (float)(state.MARINE_COOLDOWN + 1);
    float p2 = (float)(m->cooldown + 1) / (float)(state.MARINE_COOLDOWN + 1);
    float x1 = ((float)(m->pos.x) * (1 - p1) + (float)(m->direct_target->pos.x) * p1);
    float y1 = ((float)(m->pos.y) * (1 - p1) + (float)(m->direct_target->pos.y) * p1);
    float x2 = ((float)(m->pos.x) * (1 - p2) + (float)(m->direct_target->pos.x) * p2);
    float y2 = ((float)(m->pos.y) * (1 - p2) + (float)(m->direct_target->pos.y) * p2);
    gui->line((int)x1, (int)y1, (int)x2, (int)y2, col);
    if (m->direct_target->dead)
	m->direct_target = NULL;
}

void AMarine::redraw(AIGUI* gui) {
    ai_gui = gui;//
    ug->show_dmg(gui, ((80 - hp) * S_ARR_MAX) / 80);
//    gui->circle(pos, state.MARINE_SIZE, (cooldown != state.MARINE_COOLDOWN - 1) ? ClMaroon : ClRed, true);
    gui->circle(pos, state.MARINE_SIZE, (prev_hp == hp) ? ClMaroon : ClRed, true);
    if (is_enemy()) {
	cooldown++;
	draw_bullet(gui, this, ClMaroon);
    }
}

bool AMarine::is_shooting() {
    return cooldown > 6;
//    return (cooldown > state.MARINE_COOLDOWN - 3) && (cooldown < state.MARINE_COOLDOWN + 2);
}

AMarine* AMarine::nearby_enemy(int xran) {
    AMarine* e = NULL;
    int min_hp = state.MARINE_MAX_HP + 1;
    UnitGroup* g = is_ally() ? state.oppMarines : state.myMarines;
    FOREACH(g->v, m) {
	if (m->dead) // nie powinno sie zdarzyc
	    continue;
	if (dist_to(m) < range + xran)
	    if (m->hp < min_hp) {
		min_hp = m->hp;
		e = m;
	    }
    }
    return e;
}

MyMarine::MyMarine(GameObj* gob) : AMarine(gob) {
    stop();
    ignore_allies = true;
    form = NULL;
    form_index = -1;
    swapped_with = NULL;
}

void MyMarine::refresh() {
    AMarine::refresh();
    cooldown = state.gameTime - gob->component("weapon")->get_int("last_fired");
    if (direct_target)
	if (direct_target->dead)
	    direct_target = NULL;
}

void MyMarine::draw_bg(AIGUI* gui) {
    if (L2_DEBUG && (dest.x >= 0)) {
	gui->line(pos, dest, ClPink);
//	gui->line(pos, halfWay(pos, dest, M_STEP), ClBlack);
//	gui->line(pos, next_pos(M_STEP), ClGray);
    }
}

TPoint MyMarine::next_pos(int ticks) {
    return halfWay(pos, dest, ticks);
}

float angle_f(TPoint p1, TPoint p2) {
    int dx = p2.x-p1.x;
    int dy = p2.y-p1.y;
    float a = atan(dy / (dx + 0.01));
    float r = ((a / 6.28) * 32 + 32) + 16;
    if (dx < 0) r += 16;
    while (r >= 32)
	r -= 32;
    return r;
}

int angle(TPoint p1, TPoint p2) {
    return (int)(angle_f(p1, p2));
}

TPoint angle_r(TPoint p1, int a, int r) {
    a = (a + 32) % 32;
    float a1 = (a - 16) * 6.28 / 32;
    float x = (p1.x + r * cos(a1));
    float y = (p1.y + r * sin(a1));
    TPoint p2 = TPoint(x, y);
//    if (abs(angle(p1, p2) - a) > 1) {
//	cout << "ERROR: " << angle(p1, p2) << " != " << a;
//	cout << "  p=(" << p2.x - p1.x << ", " << p2.y - p1.y << ")" << endl;
//	ERR("error in AIObject::angle/angle_r");
//    }
    return p2;
}

bool aless(float a1, float a2) {
    while (a1 >= 32) a1 -= 32;
    while (a2 >= 32) a2 -= 32;
    if (a1 < 16) {
	return (a2 < a1) || (a2 > a1 + 16);
    } else {
	return (a2 < a1) && (a2 > a1 - 16);
    }
}

bool prz_odc0(TPoint p1, TPoint p2, TPoint p3, TPoint p4) {
    return (aless(angle_f(p1, p2), angle_f(p1, p3))
	!= aless(angle_f(p1, p2), angle_f(p1, p4)))
	&& (aless(angle_f(p3, p4), angle_f(p3, p1))
	!= aless(angle_f(p3, p4), angle_f(p3, p2)));
}

bool prz_odc(TPoint p1, TPoint p2, TPoint p3, TPoint p4) {
    if ((p1 == p2) && (p3 == p4))
	return false;
    if (p1 == p2)
	return prz_odc(p3, p4, p1, p2);
    int c = state.MARINE_SIZE + 4;
    if (p3 == p4) {
	if (prz_odc0(p1, p2, p3+TPoint(-c, -c), p3+TPoint(-c, c))
	|| prz_odc0(p1, p2, p3+TPoint(-c, c), p3+TPoint(c, c))
	|| prz_odc0(p1, p2, p3+TPoint(c, c), p3+TPoint(c, -c))
	|| prz_odc0(p1, p2, p3+TPoint(c, -c), p3+TPoint(-c, -c)))
	    return true;
    }
    return prz_odc0(p1, p2, p3, p4);
}

void MyMarine::redraw(AIGUI* gui) {
    AMarine::redraw(gui);
//    gui->circle(pos, state.MARINE_SIZE, (cooldown != state.MARINE_COOLDOWN-1) ? ClNavy : ClBlue, true);
//    gui->circle(pos, state.MARINE_SIZE, (prev_hp == hp) ? ClNavy : ClBlue, true);
    int h = (int)(ug->heat * 50);
    h = m_i_n(h, 255);
    gui->circle(pos, state.MARINE_SIZE, RGB(0, h, 128 - (h / 2)), true);
//    if (around) gui->circle(pos, state.MARINE_SIZE, ClGreen, true);
/*    int min_id = 0;
    FOREACH(state.myMarines->v, m)
	if (m->id > min_id)
	    min_id = m->id;
    if ((id == min_id) && (dest != nowhere)) {
	gui->line(pos, dest, ClWhite);
	FOREACH(state.myMarines->v, m)
	    if (m->id != min_id) {
		if (prz_odc(pos, dest, m->pos, m->dest)) {
		    gui->line(m->pos, m->dest, ClRed);
		} else {
		    gui->line(m->pos, m->dest, ClBlack);
		}
	    }
    }*/
    if (AI_DEBUG) {
//	if (agressive) gui->rect(pos.x-6, pos.y-6, 12, 12, ClMaroon);
        if (idle) gui->line(pos, TPoint(pos.x, pos.y-5), ClWhite);
	if (target) gui->line(pos, TPoint(pos.x+5, pos.y-5), ClGray);
        if (in_move) gui->line(pos, TPoint(pos.x+5, pos.y), ClGray);
	if (is_moving()) gui->line(pos, TPoint(pos.x+5, pos.y+5), ClGray);
        if (around) gui->line(pos, TPoint(pos.x-5, pos.y+5), ClBlue);
	if (direct_target) gui->line(pos, TPoint(pos.x+5, pos.y-5), ClGray);
        if (go_attack) gui->line(pos, TPoint(pos.x-5, pos.y), ClGray);
	if (bumps) gui->line(pos, TPoint(pos.x, pos.y+5), ClGray);
	if (old_paused) gui->rect(pos.x-7, pos.y-7, 14, 14, ClGray);
	if (swapped_with != NULL) gui->line(pos, swapped_with->pos, ClGreen);
	swapped_with = NULL;
	if (around)
	    for (int i=0; i<32; i++) {
		int n = canMove(angle_r(pos, i, M_STEP));
		gui->circle(angle_r(pos, i, 7), 1, (n > 10) ? RGB(255, m_i_n(255, n), 0) : ClBlack);
	    }
    }
    draw_bullet(gui, this, ClNavy);
    int dd = gob->dir_dmg;
    for (int i=0; i<32; i++) {
	if ((dd & (1 << i)) == 0)
	    continue;
	FOREACHTYPED(state.oppMarines->v, AMarine*, o)
	    if ((angle(pos, o->pos) == i) && (dist_to(o) <= range))
		if (o->cooldown > state.MARINE_COOLDOWN) {
		    o->cooldown = 0;
		    o->direct_target = this;
		}
    }
}

int MyMarine::dist_left() {
    TPoint d = pos;
    if (dest != nowhere)
	d = dest;
    if (target)
	d = target->pos;
    if (next_dest != nowhere)
	d = next_dest;
    return dist(pos, d);
}

void MyMarine::low_stop() {
    Vector<sint4> params;
    gob->set_action("stop", params);
    direct_dest = pos;
//    cout << id << ": low_stop\n";
}

void MyMarine::low_move(TPoint pos) {
    Vector<sint4> params;
    params.push_back(pos.x);
    params.push_back(pos.y);
    params.push_back(*gob->sod.max_speed);
    gob->set_action("move", params);
    direct_dest = pos;
//    cout << id <<  ": low_move\n";
}

void MyMarine::low_attack(AMarine* aiob) {
    if (dist_to(aiob) > range) {
	cerr << "BLAD: cel poza zasiegiem!\n";
	return;
    }
    Vector<sint4> params;
    params.push_back(state.gsm->get_game().get_cplayer_info().get_id(aiob->gob));
//    cout << id << ": low_attack" << endl;
    gob->component("weapon")->set_action("attack", params);
    direct_target = aiob;
}

float MyMarine::canMove0(TPoint where) {
    float d = dist_f(pos, where) + 0.01;
    if (d > 2 * M_STEP)
	d = 2 * M_STEP;
    float min_d = 1000;
    float tmp_d;
    TPoint hw[2 * M_STEP+1];
    for (int d1=0; d1<=d; d1++)
	hw[d1] = halfWay(pos, where, d1);
    FOREACHTYPED(state.allUnits->v, AIObject*, aiob) {
	if (dist_f(pos, aiob->pos) > 25)
	    continue;
	if ((aiob->id == id) || (aiob->dead))
	    continue;
	if (ignore_allies && aiob->is_ally() && (!aiob->is_shooting())) {
	    if ((tmp_d = dist(((MyMarine*)aiob)->dest, where)) < min_d)
		min_d = tmp_d;
	    continue;
	}
	for (int d1=2; d1<=d; d1++) {
	    if (aiob->is_moving())
		if ((tmp_d = dist_f(aiob->next_pos(d1), hw[d1])) <= min_d)
		    min_d = tmp_d;
	    if ((tmp_d = dist_f(aiob->pos, hw[d1])) <= min_d) {
//		cout << id << " -- canMove: " << d1 << " (" << (int)(aiob->pos.x) << ", " << (int)(aiob->pos.y) << ")->(" << (int)(hw[d1].x) << ", " << (int)(hw[d1].y) << ") = " << (int)tmp_d << endl;
		min_d = tmp_d;
	    }
	}
    }
    return min_d;
}

int MyMarine::canMove(TPoint where) {
    if ((where.x < 4) || (where.y < 4)
	|| (where.x > state.MAP_WIDTH - 4)
	|| (where.y > state.MAP_HEIGHT - 4))
	    return 0;
    int d0 = 1000;
    FOREACHTYPED(state.allUnits->v, AIObject*, aiob) {
	if ((aiob->id == id) || (aiob->dead))
	    continue;
	if (dist_f(pos, aiob->pos) < d0)
	    d0 = dist_f(pos, aiob->pos);
    }
    return (int)((canMove0(where) - d0) * 10);
}

TPoint MyMarine::OnBeforeCollision() {
    if (around && (is_moving()))
	return nowhere;
    TPoint p;
    int a = angle(pos, dest);
    for (int i=1; i<15; i++) {
	p = angle_r(pos, a+i, M_STEP);
	if (canMove(p) > 10)
	    return p;
	p = angle_r(pos, a-i, M_STEP);
	if (canMove(p) > 10)
	    return p;
    }
    return dest;
//    return pos + TPoint(rand() % 5 - 2, rand() % 5 - 2);
/*    for (int i=0; i<10; i++) {
	p = pos + TPoint(rand() % 10 - 5, rand() % 10 - 5);
	int d = canMove(p);
	if (d > 12 + rand() % 3)
	    return p;
    }
    return p;*/
}

void MyMarine::update() {
    old_paused = paused;
    if (paused) {
	paused = false;
	return;
    }
    TPoint tmp_dest = nowhere;
    AMarine* tmp_target = NULL;
    if (direct_target)
	if (direct_target->dead)
	    direct_target = NULL;
    if (in_move) {
	if (target && !around) {
	    if (target->dead) {
		target = NULL;
		cout << id << " stopped: target died" << endl;
		stop();
		return;
	    }
	    if (is_moving()) {
		dest = target->pos;
		tmp_dest = dest;
	    }
	}
	if ((pos == dest) && (!is_moving()) && (!target) && (!direct_target)) {
	    cout << id << " stopped: reached dest." << endl;
	    stop();
	}
	if ((!is_moving()) && (!direct_target))
	    tot_bumps++;
	if (tot_bumps > max_bumps) {
	    cout << id << " stopped: too many colisions" << endl;
	    stop();
	    return;
	}
	int d = dist_left();
	if (d > M_STEP)
	    d = M_STEP;
	if (!(direct_target && is_shooting())) {
	    if (canMove0(next_pos(d)) <= 12) {
//	    	cout << id << ": around" << endl;
		tmp_dest = OnBeforeCollision();
		around = true;
	    }
	    if (canMove0(halfWay(pos, dest, d)) > 12) {
//	    	cout << id << ": back" << endl;
		around = false;
		tmp_dest = dest;
	    }
	}
	if (go_attack && target && is_shooting())
	    if (dist_to(target) <= range)
		tmp_target = (AMarine*)target;
    }
    //?
    AMarine* t2;
    if (stopped && (!nearby_enemy()) && (t2 = nearby_enemy(30))) {
	if ((next_dest == nowhere) && !in_move)
	    next_dest = pos;
	low_move(t2->pos);
	return;
    }
    if (stopped && (next_dest != nowhere) && (next_dest != pos)) {
	in_move = true;
	dest = next_dest;
	next_dest = nowhere;
	low_move(dest);
	return;
    }
    ///
    if (!tmp_target && agressive)
	tmp_target = nearby_enemy();
    if (tmp_target/* && is_shooting()*/) {
	if (is_moving()) { //?
	    low_stop();
//	    cout << id << ": cooldown=" << cooldown << " (stop)" << endl;
	} else {
	    low_attack(tmp_target);
//	    cout << id << ": cooldown=" << cooldown << " (attack)" << endl;
	}
    } else {
	if (tmp_dest != nowhere)
	    low_move(tmp_dest);
    }
}

void MyMarine::wait() {
    low_stop();
    paused = true;
}

void MyMarine::stop() {
    direct_target = NULL;
    prev_dist = -1;
    bumps = 0;
    tot_bumps = 0;
    max_bumps = 100; // default
    paused = false;
    around = false;
    in_move = false;
    go_attack = false;
    agressive = true; // changed
    paused = false;
    old_paused = false;
    stopped = true;
    idle = true;
    dest = nowhere;
    next_dest = nowhere;
    target = NULL;
    low_stop();
}

void MyMarine::moveTo(TPoint p) {
//    direct_target = NULL;
    go_attack = false;
    target = NULL;
    bumps = 0;
//    stop();
    idle = false;
    in_move = true;
    stopped = false;
    agressive = false; // changed
    dest = p;
//    prev_dist = dist(pos, dest) + 1;
//    low_move(dest);
}

void MyMarine::follow(AIObject* aiob) {
    moveTo(aiob->pos);
    target = aiob;
}

void MyMarine::attack(AMarine* aiob) {
    if (!aiob->is_enemy())
	ERR("Cannot attack allies nor neutrals");
    if (aiob->dead)
	ERR("Cannot attack dead unit");
    follow(aiob);
    go_attack = true;
}

void MyMarine::holdPosition() {
    stop();
    stopped = false;
}

void MyMarine::autoAttack(TPoint p) {
    moveTo(p);
    agressive = true;
}

void MyMarine::autoAttack(AMarine* aiob) {
    follow(aiob);
    agressive = true;
}
