#ifndef AIObject_H
#define AIObject_H

#include "Common.H"
#include "GameObj.H"
#include "AIGUI.H"

#define PPOS_LEN 5
#define M_STEP 5

class AIObject {
    protected:
	AIGUI* ai_gui; // delete
    public:
	GameObj* gob;
	int id;
	bool dead;
	int owner;
	TPoint pos;
	TPoint prev_pos[PPOS_LEN];
	AIObject(GameObj* gob);
	Unit_gui* ug;
	virtual ~AIObject() {};
	virtual bool is_ally() {return false;};
	virtual bool is_enemy() {return false;};
	virtual void refresh();
	virtual void draw_bg(AIGUI* gui) {gui->circle(pos, 0, ClBlack);}; // unused parameter
	virtual void redraw(AIGUI* gui);
	virtual bool is_shooting() {return false;};
	bool is_moving();
	float dist_to(AIObject* aiob);
	virtual TPoint next_pos(int ticks);
};

class Asheep : public AIObject {
    public:
	Asheep(GameObj* gob);
	void redraw(AIGUI* gui);
};

class AMarine : public AIObject {
    public:
	int range;
	int hp;
	int prev_hp;
	int cooldown;
	AMarine* direct_target;
	AMarine(GameObj* gob);
	bool is_enemy() {return true;};
	void refresh();
	void redraw(AIGUI* gui);
	bool is_shooting();
	AMarine* nearby_enemy(int xran = 0);
};

class MyMarine : public AMarine {
    private:
	bool go_attack;			// rozkaz ataku konkretnej jednostki
	int prev_dist;
	int bumps;			// liczba ostatnich zderzen
	int tot_bumps;			// liczba zderzen od wydania poprzedniego rozkazu
	bool around;			// omijanie przeszkody
	TPoint next_dest;		// kolejny cel
	AIObject* target;		// bierzacy cel (przeciwnik)
	bool in_move;
	bool agressive;			// atakuj wszystkich po drodze
	bool stopped;			// gonic wszystkich prz. ktorzy sie zbliza
	bool paused;			// zatrzymana na 1 klatke
	bool old_paused;
	bool ignore_allies;
	float canMove0(TPoint where);
	TPoint OnBeforeCollision();
    public:
	void* form; // Formation*
	int form_index;
	MyMarine* swapped_with;
	TPoint dest;			// bierzacy cel (punkt)
	TPoint direct_dest;
	int max_bumps;			// po ilu kolizjach sie zatrzymuje
	bool idle;			// objekt ma rozkaz "stop" lub "holdPosition"
	MyMarine(GameObj* gob);
	bool is_ally() {return true;};
	void refresh();			// aktualizuje stan jednostki
	void draw_bg(AIGUI* gui);	// rysuje spodnia warstwe
	void redraw(AIGUI* gui);	// rysuje wierzchnia warstwe
	int dist_left();		// ile pixeli zostalo do przejscia
	TPoint next_pos(int ticks);
	int canMove(TPoint where);
	void low_stop();
	void low_move(TPoint pos);
	void low_attack(AMarine* aiob);
	void update();			// aktualizuje niskopoziomowe rozkazy
					// (np. po kolizji, przypomina jednostce, co miala zrobic)
	void wait();			// zatrzymuje sie na chwile
	void stop();			// przerywa wykonywanie bierzacej akcji (ilde:=true)
					// goni przeciwnikow, ktorzy sie zbliza do niego
	void moveTo(TPoint p);		// idzie do punktu p (omijajac przeszkody)
	void follow(AIObject* aiob);	// idzie w strone danej jednostki (samonaprowadzajace)
	void attack(AMarine* aiob);	// idzie zaatakowac dana jednostke
	void holdPosition();		// utrzymuje bierzaca pozycje, atakujac przeciwnika tylko
					// gdy znajdzie sie w zasiegu ataku
	void autoAttack(TPoint p);	// idze w strone punktu p i atakuje wszystkich po drodze
	void autoAttack(AMarine* aiob);	// jw. ale idzie w strone podanej jednostki (follow)
};

bool prz_odc(TPoint p1, TPoint p2, TPoint p3l, TPoint p4);
TPoint halfWay(TPoint pos1, TPoint pos2, int d);
TPoint angle_r(TPoint p1, int a, int r);
int angle(TPoint p1, TPoint p2);

#endif
