#include <iostream.h>

#include "GameChanges.H"
#include "GameObj.H"

#include "AIState.H"
#include "AIObject.H"
#include "UnitGroup.H"

#include "Game.H"

#define chk_gob() \
    gob = (*it)->get_GameObj();\
    if (!gob) continue;\
    if (!gob->sod.in_game) continue;

void InitAI() {
    cout << "***************** I N I T I A L I Z I N G . . . *******************" << endl;
    state.initialized = true;
    state.last_aiob_id = 0;
    state.quit = false;
    state.SHEEP_SIZE = -1;
    state.MARINE_SIZE = -1;
    int tilep = state.gsm->get_game().get_tile_points();
    state.MAP_WIDTH = state.gsm->get_game().get_map().get_width() * tilep;
    state.MAP_HEIGHT = state.gsm->get_game().get_map().get_height() * tilep;
    state.gameTime = 0;
    state.playerId = state.gsm->get_game().get_client_player();
    state.allUnits = new UnitGroup();
    state.sheeps = new UnitGroup();
    state.myMarines = new UnitGroup();
    state.oppMarines = new UnitGroup();
    state.curr_target = NULL;
    state.AIstage = 0;
    if (state.gui)
	cout << "2d gui enabled" << endl;
}

void AddUnit(GameObj* gob) {
    AIObject* aiob;
    int o = *gob->sod.owner;
    if (o == 2) {
	aiob = new Asheep(gob);
	state.sheeps->AddUnit(aiob);
    } else {
	if (o == state.playerId) {
	    aiob = new MyMarine(gob);
	    state.myMarines->AddUnit(aiob);
	} else {
	    aiob = new AMarine(gob);
	    state.oppMarines->AddUnit(aiob);
	}
    }
    state.allUnits->AddUnit(aiob);
}

void RemoveUnit(GameObj* gob) {
    state.allUnits->RemoveUnit(gob);
    int o = *gob->sod.owner;
    if (o == 2) {
	state.sheeps->RemoveUnit(gob);
    } else {
	if (o == state.playerId) {
	    state.myMarines->RemoveUnit(gob);
	} else {
	    state.oppMarines->RemoveUnit(gob);
	}
    }
}

void UpdateWorld() {
    state.gameTime = state.gsm->get_game().tick(); //++
    cout << "---------------------------------- Frame nr " << state.gameTime << " ----------------------------------\n";
    GameObj* gob;
    const GameChanges* changes = &state.gsm->get_changes();
    FORALL(changes->new_objs, it) {
	chk_gob();
	if ((state.SHEEP_SIZE == -1) && (*gob->sod.owner == 2)) {
	    state.SHEEP_SIZE = *gob->sod.radius;
	    cout << "sheep radius: " << state.SHEEP_SIZE << endl;
	}
	if ((state.MARINE_SIZE == -1) && (*gob->sod.owner < 2)) {
	    state.MARINE_SIZE = *gob->sod.radius;
	    state.MARINE_COOLDOWN = gob->component("weapon")->get_int("cooldown");
	    state.MARINE_MAX_HP = gob->get_int("hp");
	    state.MARINE_RANGE = gob->component("weapon")->get_int("max_ground_range");
	    cout << "marine radius: " << state.MARINE_SIZE << endl;
	}
	if (state.hcap && (*gob->sod.owner == state.playerId)) {
	    state.hcap--;
	    Asheep* sheep = new Asheep(gob);
	    state.sheeps->AddUnit(sheep);//?
	    continue;
	}
	AddUnit(gob);
    }
    FORALL(changes->vanished_objs, it) {
	chk_gob();
	RemoveUnit(gob);
    }
    FORALL(changes->dead_objs, it) {
	chk_gob();
	RemoveUnit(gob);
    }
    FORALL(state.allUnits->v, it)
	(*it)->refresh();
}

void ComputeActions0() { // stata wersja
    state.myMarines->autoAttack((AMarine*)state.oppMarines->any());
    state.myMarines->update();
}

void ComputeActions1() {
   if (state.AIstage == 0) {						// inicjalizacja
	state.f1 = new Formation();
	FOREACH(state.myMarines->v, it)
	    state.f1->AddUnit(it);
	for (int i=0; i<MAX_MARINES; i++)
//	    state.f1->shape[i] = TPoint((i%1)*0, (i-25)*10);
	    state.f1->shape[i] = TPoint((i%2)*40-20, (i/2-12)*18);
	state.f1->init(state.myMarines->middle());
	state.AIstage = 1;
    } else if (state.AIstage == 1) {					// tworzenie formacji
	if ((state.f1->busy() < 3) || (state.f1->underAttack()))
	    state.AIstage = 2;
    } else if (state.AIstage == 2) {					// pelny atak
	state.f1->autoAttack(state.oppMarines->middle());
	if (dist(state.f1->middle(), state.oppMarines->middle()) < 30)
	    state.AIstage = 3;
    } else if (state.AIstage == 3) {								// rozwalanie niedobitkow
	if (!state.curr_target || state.curr_target->dead)
	    state.curr_target = (AMarine*)state.oppMarines->any();
	state.f1->autoAttack(state.curr_target);
    }
    cout << "### stage " << state.AIstage << endl;
    state.f1->update();
}

bool ignore[200];

TPoint Edist(TPoint p) {
    float min = 2000000000;
    float min2 = 2000000000;
    FOREACH(state.oppMarines->v, o) {
	if (!ignore[o->id]) {//?
	    float d = dist_k(p, o->pos);
	    if (d < min) {
		min2 = min;
		min = d;
	    } else if (d < min2) {
		min2 = d;
	    }
	}
    }
    min = sqrt(min);
    min2 = sqrt(min2);
    return TPoint(min, min-min2);
}

TPoint Adist(TPoint p) {	// copy & paste
    float min = 2000000000;
    float min2 = 2000000000;
    FOREACH(state.myMarines->v, o) {
	float d = dist_k(p, o->pos);
	if (d < min) {
	    min2 = min;
	    min = d;
	} else if (d < min2) {
	    min2 = d;
	}
    }
    min = sqrt(min);
    min2 = sqrt(min2);
    return TPoint(min, min-min2);
}

int AInRange(TPoint p, float r) {
    int n = 0;
    r *= r;
    FOREACH(state.myMarines->v, m)
	if (dist_k(p, m->pos) <= r)
	    n++;
    return n;
}

int EInRange(TPoint p, float r) {
    int n = 0;
    r *= r;
    FOREACH(state.oppMarines->v, m)
	if (dist_k(p, m->pos) <= r)
	    n++;
    return n;
}

// 3000 12 2 15 250 10 3 50 3 100
#define wallp			5000
#define speed			12
#define eval_precision		2
#define eval_dist		15
#define eval_long		250
#define expand_front_line 	10
#define auto_formation		3
#define can_shoot_bonus		0
#define too_close_penalty	3
#define close_ally_penalty	10
#define avoid_enemy		10000
#define avoid_range		1

void ComputeActions() {
    for (int i=0; i<200; i++)
	ignore[i] = false;
    if (state.oppMarines->size > 10)
    FOREACH(state.oppMarines->v, o)
	if (EInRange(o->pos, 50) < 3)
	    if (AInRange(o->pos, state.MARINE_RANGE+5) == 0)
		ignore[o->id] = true;
    float avd = 0;
    FOREACH(state.myMarines->v, m) {
	float d = Edist(m->pos).x;
	avd += d / state.myMarines->size;
    }
    avd -= speed;
    avd -= state.gameTime / 100;
    avd = m_a_x(avd, state.MARINE_RANGE - 5);
    int ec[200];
    float fp[200];
    FOREACH(state.oppMarines->v, o) {
	ec[o->id] = 0;
	FOREACH(state.myMarines->v, m)
	    if (m->dist_to(o) <= state.MARINE_RANGE)
		ec[o->id]++;
	fp[o->id] = (avoid_enemy / avoid_range) * (1 / (ec[o->id] + 0.1));
    }
    FOREACH(state.myMarines->v, m) {
	m->ug->heat = 0;
	float bval = -1000000;
	TPoint bp = m->pos;
	for (int i=0; i<32; i+=eval_precision) {
	    TPoint np = angle_r(m->pos, i, eval_dist);
	    float r2 = Edist(m->pos).x / 2;
	    r2 = m_i_n(m_i_n(r2, eval_long), m_i_n(m->pos.x, m->pos.y));
	    r2 = m_i_n(r2, m_i_n(state.MAP_WIDTH - m->pos.x, state.MAP_HEIGHT - m->pos.y));
	    TPoint nnp = angle_r(m->pos, i, r2);
		if (!m->canMove(np))
		    continue;
		TPoint enp = Edist(np);
		TPoint em = Edist(m->pos);
		float val = -abs(enp.x - avd) * expand_front_line;
		if (enp.x < m->range) {
		    if (EInRange(np, 120) < AInRange(np, 120)) {
			val += can_shoot_bonus;
		    } else {
			val -= can_shoot_bonus;
		    }
		}
		if (em.x < m->range)
		    val += (em.x - m->range) * too_close_penalty;
		val -= Edist(nnp).y * auto_formation;
		FOREACH(state.myMarines->v, m2)
		    if (m->id != m2->id)
			val -= close_ally_penalty / dist_f(m->pos, m2->pos);
		val -= (wallp / m->pos.x) + (wallp / m->pos.y)
		    + (wallp / (state.MAP_WIDTH-m->pos.x))
		    + (wallp / (state.MAP_HEIGHT-m->pos.y));
		bool can_attack = false;
		float dt[200];
		FOREACH(state.oppMarines->v, o) {
		    dt[o->id] = m->dist_to(o);
		    if (dt[o->id] <= state.MARINE_RANGE)
			can_attack = true;
		}
		float heat = 0;
		for (int j=0; j<2*avoid_range; j+= 2)
		    FOREACH(state.oppMarines->v, o) {
			if (dt[o->id] <= state.MARINE_RANGE+j) {
			    val -= fp[o->id];
			    heat += fp[o->id] / avoid_enemy;
			}
		    }
		if (can_attack) {
		    val += avoid_enemy;
		    heat -= 1;
		}
		if (val > bval) {
		    bval = val;
		    bp = np;
		    m->ug->heat = m_a_x(0, heat);
		}
	    }
	if ((m->cooldown < 6) || (m->ug->heat > 250)) {
	    m->moveTo(bp);
	} else {
	    m->autoAttack(bp);
	}
    }
    state.myMarines->update();
}
