#include "Common.H"
#include "Formation.H"
#include "AIState.H"
#include "AIObject.H"

Formation::Formation() : UnitGroup() {
    pos = TPoint(-1, -1);
    for (int i=0; i<MAX_MARINES; i++)
	shape[i] = TPoint(0, 0);
    for (int i=0; i<200; i++)
	idToPos[i] = -1;
    last_ua = 1000;
    stop();
}

void Formation::AddUnit(AIObject* aiob) {
    UnitGroup::AddUnit(aiob);
    MyMarine* m = (MyMarine*)aiob;
    m->form = this;
    m->form_index = size - 1;
}

float Formation::dist_left1(MyMarine* m) {
    return dist_f(m->pos, single_dest + shape[idToPos[m->id]]);
}

TPoint Formation::temp_dest(MyMarine* m) {
    return halfWay(m->pos, single_dest + shape[idToPos[m->id]], dist_left1(m) * 0.2);
}

int Formation::dist_left() {
    int max = 0;
    int d;
    FOREACH(v, m)
	if ((d = dist_left1(m)) > max)
//	if ((d = m->dist_left()) > max)
	    max = d;
    return max;
}


bool Formation::underAttack() {
    if (UnitGroup::underAttack()) {
	last_ua = 0;
    } else {
	last_ua++;
    }
    return last_ua < state.MARINE_COOLDOWN;
}

void Formation::init(TPoint where) {
    pos = where;
    bool used[MAX_MARINES];
    for (int i=0; i<MAX_MARINES; i++)
	used[i] = false;
    int d, j;
    FOREACH(v, m) {
	int min = DUZO;
	for (int i=0; i<MAX_MARINES; i++)
	    if (!used[i] && (d = dist(m->pos, where+shape[i])) < min) {
		min = d;
		j = i;
	    }
	idToPos[m->id] = j;
	m->moveTo(where + shape[j]);
	used[j] = true;
    }
}

void Formation::swap(int i1, int i2) {
    MyMarine* m1 = (MyMarine*)(v[i1]);
    MyMarine* m2 = (MyMarine*)(v[i2]);
//    v[i1] = v[i2];
//    v[i2] = m1;
    int i = idToPos[m1->id];
    idToPos[m1->id] = idToPos[m2->id];
    idToPos[m2->id] = i;
    TPoint p;
    p = m1->dest;
    m1->dest = m2->dest;
    m2->dest = p;
    p = m1->direct_dest;
    m1->direct_dest = m2->direct_dest;
    m2->direct_dest = p;
    m1->swapped_with = m2;
    m2->swapped_with = m1;
    m1->low_move(m1->direct_dest);
    m2->low_move(m2->direct_dest);
}

void Formation::update() {
    if (order == ORDER_AAM)
	single_dest = single_target->pos;
    if (underAttack()) {
	if (single_dest != nowhere)
	    UnitGroup::autoAttack(single_dest);
    } else {
	if (order == ORDER_AAM)
	    autoAttack(single_dest);
	if (single_dest != nowhere) {
	FOREACH(v, m)
	    m->autoAttack(temp_dest(m));
            int d = dist_left();
	    FOREACH(v, m)
		if (dist_left1(m) < d * 0.9) // dlaczego 0.9?
		    m->wait();
	}
    }
    UnitGroup::update();
    for (int i1=0; i1<size; i1++)
	for (int i2=i1+1; i2<size; i2++) {
	    MyMarine* m1 = (MyMarine*)(v[i1]);
	    MyMarine* m2 = (MyMarine*)(v[i2]);
	    if ((m1->dest == nowhere) || (m2->dest == nowhere))
		continue;
	    if (m_a_x(dist_f(m1->pos, m2->dest), dist_f(m2->pos, m1->dest))
		< (m_a_x(dist_f(m1->pos, m1->dest), dist_f(m2->pos, m2->dest))))
		    swap(i1, i2);
/*	    if (prz_odc(m1->pos, m1->dest, m2->pos, m2->dest)
		&& !prz_odc(m1->pos, m2->dest, m2->pos, m2->dest))
		    swap(i1, i2);
		{    bool ok = true;
		    for (int i3=0; i3<size; i3++)
			if ((i3 != i1) && (i3 != i2)) {
			    MyMarine* m3 = (MyMarine*)(v[i3]);
			    if (prz_odc(m1->pos, m3->dest, m3->pos, m1->dest))
				ok = false;
			}
		    if (ok) swap(i1, i2);
		}*/
	}
}

void Formation::stop() {
    UnitGroup::stop();
    single_target = NULL;
    single_dest = nowhere;
    order = ORDER_STOP;
}

void Formation::holdPosition() {
    UnitGroup::holdPosition();
    single_target = NULL;
    single_dest = nowhere;
}

void Formation::moveTo(TPoint p) {
    int i=0;
    FOREACH(v, m) {
	m->moveTo(p + shape[idToPos[m->id]]);
	i++;
    }
    single_dest = p;
}

void Formation::autoAttack(TPoint p) {
    int i=0;
    FOREACH(v, m) {
	m->autoAttack(p + shape[idToPos[m->id]]);
	i++;
    }
    single_dest = p;
    order = ORDER_AAP;
}

void Formation::autoAttack(AMarine* aiob) {
    autoAttack(aiob->pos);
    single_dest = nowhere;
    single_target = aiob;
    order = ORDER_AAM;
}
