#include <iostream>
#include <sys/time.h>

#include "GameObj.H"
#include "GUI.H"

#include "Common.H"
#include "TPoint.H"
#include "AIState.H"
#include "Graph.H"
#include "L1.H"
#include "L2.H"
#include "L3.H"
#include "UnitGroup.H"
#include "AIGUI.H"

using namespace std;

AIGUI::AIGUI() {
    GUI::GUI();
    for (int i=0; i<MAX_UNITS; i++)
	for (int j=0; j<SMOKE_LENGTH; j++)
	    for (int k=0; k<3; k++)
		s_arr[i][j][k] = -1;
    for (int i=0; i<MAXPROFILERS; i++)
	clearTime(i);
}

void AIGUI::initGui() {
    init(state.params.window_width, state.params.window_height, state.gsm->get_game());
    int gt = state.GAME_TYPE;
    adjust_view_mag((gt == 1) ? 0 : ((gt == 4) ? -2 : -6));
    c_notice << "GUI initialized" << endl;
}

int AIGUI::currTime() {
    struct timeval tv;
    gettimeofday(&tv, 0);
    int time = 10000 * (tv.tv_sec % (MAXINT/10000)) + tv.tv_usec / 100;
    c_debug << "time is: " << time << endl;
    return time;
}

void AIGUI::clearTime(int id) {
    c_debug << "clearTime(" << id << ")" << endl;
    RCHK(id, MAXPROFILERS);
    starts[id] = times[id] = currTime();
    colors[id] = ClBlack;
    for (int i=0; i<MAXPROFILETIME; i++)
	msecs[id][i] = ClBlack;
}

void AIGUI::timeStart(int id, const RGB& color) {
    RCHK(id, MAXPROFILERS);
    int ct = timeStop(id);
    times[id] = ct;
    colors[id] = color;
}

int AIGUI::timeStop(int id) {
    RCHK(id, MAXPROFILERS);
    int ct = currTime() - starts[id];
    if (ct >= MAXPROFILETIME)
	ct = MAXPROFILETIME-1;
    for (int i=times[id]; i<=ct; i++)
	msecs[id][i] = colors[id];
    times[id] = ct;
    colors[id] = ClBlack;
    return ct;
}

void AIGUI::showProfiler(int id, int x1, int y1, int x2, int y2) {
    for (int i=x1; i<x2; i++)
	line(i, y1, i, y2, msecs[id][(i-x1)*(MAXPROFILETIME-1)/(x2-x1)]);
    if (MAXPROFILETIME > TIMEREDLINE) {
	int x = x1+TIMEREDLINE*(x2-x1)/(MAXPROFILETIME-1);
	line(x, y1-5, x, y2+5, ClRed);
    }
    rect(x1, y1, x2, y2, ClWhite);
}

void AIGUI::beforeDisplay() {
    if (SDL_MUSTLOCK(screen) && (SDL_LockSurface(screen) < 0)) {
	c_error << "Can't lock screen: \"" << SDL_GetError() << "\"" << endl;
	return;
    }
    correct_view();
    clear_screen();
}

void AIGUI::afterDisplay() {
    if (SDL_MUSTLOCK(screen))
	SDL_UnlockSurface(screen);
    SDL_UpdateRect(screen, 0, 0, 0, 0);
}

void AIGUI::display() {
    for (int l=0; l<2; l++) {
	if (state.GAME_TYPE == 1) {
	    drawGame1(l);
	} else if (state.GAME_TYPE == 2) {
	    drawGame2(l);
	} else if (state.GAME_TYPE == 4) {
	    drawGame4(l);
	} else {
	    ERR("AIGUI: unknown game scenario");
	}
    }
    drawInt(40, state.MAP_HEIGHT - 30, 10, 20, state.game_time, ClBlack);
    if (state.params.show_profilers)
	showProfiler(2, mapW-120, mapH-16, mapW-20, mapH-10);
    if (state.old_behind || (state.behind > 0)) {
	TPoint p(mapW-50, mapH-50);
	circle(p, 28, ClRed, true);
	circle(p, 25, ClYellow, true);
	line(p, p + TPoint(0, -22), ClNavy);
	line(p, p + TPoint(18, 0), ClNavy);
    }
}

void AIGUI::line(int x1, int y1, int x2, int y2, const RGB& c) {
    draw_line(views[0], x1, y1, x2, y2, c);
}

void AIGUI::line(TPoint p1, TPoint p2, const RGB& c) {
    draw_line(views[0], (int)(p1.x), (int)(p1.y), (int)(p2.x), (int)(p2.y), c);
}

void AIGUI::line(int node1, int node2, const RGB& c) {
    line(state.map->nodePos(node1), state.map->nodePos(node2), c);
}

void AIGUI::circle(int x, int y, int r, const RGB& c, bool filled) {
    draw_circle(views[0], x, y, r, c, filled);
}

void AIGUI::circle(TPoint s, int r, const RGB& c, bool filled) {
    draw_circle(views[0], (int)(s.x), (int)(s.y), r, c, filled);
}

void AIGUI::circle(int node, int r, const RGB& c, bool filled) {
    circle(state.map->nodePos(node), r, c, filled);
}

void AIGUI::putPixel(TPoint p, const RGB& c) {
    draw_circle(views[0], (int)(p.x), (int)(p.y), 0, c, true);
}

void AIGUI::filledRect(int x1, int y1, int x2, int y2, const RGB& c) {
    for (int i=y1; i<=y2; i++)
	line(x1, i, x2, i, c);
}

void AIGUI::rect(int x1, int y1, int x2, int y2, const RGB& c) {
    line(x1, y1, x2, y1, c);
    line(x2, y1, x2, y2, c);
    line(x2, y2, x1, y2, c);
    line(x1, y2, x1, y1, c);
}

void AIGUI::mark(Hardware* h, int type) {
    int arr[8][2] = {
	{0, -5}, {5, -5}, {5, 0}, {5, 5},
	{0, 5}, {-5, 5}, {-5, 0}, {-5, -5}};
    RGB colors[8] = {ClWhite, ClGray, ClDarkGreen, ClGray, ClGray, ClGray, ClGray, ClGray};
    line(h->pos, h->pos + TPoint(arr[type][0], arr[type][1]), colors[type]);
}

void AIGUI::clear_screen() {
    filledRect(0, 0, mapW, mapH, ClBackground);
    rect(0, 0, mapW, mapH, ClGray);
}

void AIGUI::drawDigth(int x, int y, int w, int h, int n, const RGB& c) {
    int dtab[10][7] = {
	{1,1,1,1,1,1,0},
	{0,1,1,0,0,0,0},
	{1,1,0,1,1,0,1},
	{1,1,1,1,0,0,1},
	{0,1,1,0,0,1,1},
	{1,0,1,1,0,1,1},
	{1,0,1,1,1,1,1},
	{1,1,1,0,0,0,0},
	{1,1,1,1,1,1,1},
	{1,1,1,1,0,1,1}
    };
    int dpos[7][4] = {
	{0,0,1,0},
	{1,0,1,1},
	{1,1,1,2},
	{0,2,1,2},
	{0,1,0,2},
	{0,0,0,1},
	{0,1,1,1}
    };
    h /= 2;
    for (int i=0; i<7; i++)
	if (dtab[n][i])
	    line(x+dpos[i][0]*w, y+dpos[i][1]*h, x+dpos[i][2]*w, y+dpos[i][3]*h, c);
}

void AIGUI::drawInt(int x, int y, int w0, int h, int n, const RGB& c) {
    bool minus = n < 0;
    if (minus)
	n = -n;
    do {
	drawDigth(x, y, w0*4/5-1, h, n%10, c);
	n /= 10;
	x -= w0;
    } while (n > 0);
    if (minus)
	line(x-w0, y+h/2, x-w0/5+2, y+h/2, c);
}

void AIGUI::drawHp(UnitGroup* units, int x1, int y1, int x2, int y2, const RGB& fc, const RGB& sc) {
    int l = units->totalLife(0);
    int o = units->totalLife(HP_HISTORY_LENGTH-1);
    int n = MAX(0, 2*l-o);
    int m = units->maxLife() + 1;
    filledRect(x1, y1, x1 + (x2 - x1) * l / m, y2, sc);
    filledRect(x1, y1, x1 + (x2 - x1) * n / m, y2, fc);
    rect(x1, y1, x2, y2, ClGray);
}

int rand30(int n) {
    int r = n - rand() % 30;
    return MAX(0, r);
}

void AIGUI::showDamage(AIObject* aiob, int n, TPoint dpos) {
    int m = MAX(0, SMOKE_LENGTH-n-1);
    int id = aiob->hardware->id;
    RCHK(id, MAX_UNITS);
    RCHK(m, SMOKE_LENGTH);
    s_arr[id][m][0] = (int)(aiob->hardware->pos.x + dpos.x);
    s_arr[id][m][1] = (int)(aiob->hardware->pos.y + dpos.y);
    s_arr[id][m][2] = m + rand() % (m+1);
    if (aiob->hardware->dead)
	for (int i=0; i<=m; i++)
	    s_arr[id][i][0] = -1;
    for (int k=0; k<4; k++)
	for (int i=SMOKE_LENGTH-1; i>m; i--)
	    for (int j=0; j<3; j++)
		s_arr[id][i][j] = s_arr[id][i-1][j];
    for (int i=0; i<SMOKE_LENGTH; i++)
	if (s_arr[id][i][0] >= 0) {
	    s_arr[id][i][0] += rand() % 5 - 2;
	    if (rand() % (s_arr[id][i][2] + 1) < 25)
		s_arr[id][i][1] -= rand() % 4 - 1;
	    s_arr[id][i][2] += rand() % 3;
	}
    for (int i=0; i<SMOKE_LENGTH; i++)
        if (s_arr[id][i][0] >= 0)
    	    circle(s_arr[id][i][0], s_arr[id][i][1], 1,
		RGB(rand30(s_arr[id][i][2]),
		    rand30(s_arr[id][i][2]),
		    rand30(s_arr[id][i][2])));
}

void AIGUI::showUnit(AIObject* aiob, const RGB& c1, const RGB& c2, int layer) {
    Hardware* hw = aiob->hardware;
    if (layer == 0) {
	if ((hw->cooldown > 0) && (hw->target)) {
	    TPoint p1 = halfWay(hw->pos, hw->target->pos,
		1-((float)(hw->cooldown) / (hw->MAX_COOLDOWN+1)));
	    TPoint p2 = halfWay(hw->pos, hw->target->pos,
		1-((float)(hw->cooldown+1) / (hw->MAX_COOLDOWN+1)));
	    line(p1, p2, c1);
	}
    } else if (layer == 1) {
	if (!hw->INVINCIBLE) {
	    int n = (hw->MAX_HP - hw->hp) * SMOKE_LENGTH / hw->MAX_HP;
	    showDamage(aiob, n);
	}
	RGB c = (hw->hp == hw->old_hp[1]) ? c1 : c2;
	if (aiob->hardware->HW_TYPE == HW_WORKER2)
	    c = c2;
	circle(hw->pos, MAX(hw->SIZE, 3), c, true);
	if ((state.params.show_dest) && (aiob->owner == state.player_id)
	    && (aiob->software->low_dest != nowhere))
	    line(aiob->hardware->pos, aiob->software->low_dest, ClGray);
	if (state.params.show_unit_state)
	    if (aiob->owner == state.player_id) {
		if (state.params.show_path)
		    line(aiob->hardware->pos, state.map->nodePos(aiob->software->next_node), ClBlack);
		if (aiob->software->ready)
		    mark(aiob->hardware, 0);
		if (aiob->software->init)
		    mark(aiob->hardware, 1);
		if (aiob->hardware->HW_TYPE == HW_WORKER2)
		    mark(aiob->hardware, 2);
		if (aiob->software->ignore)
		    mark(aiob->hardware, 3);
		if (aiob->software->paused)
		    mark(aiob->hardware, 4);
		if (aiob->software->mode == MODE_SYNC)
		    mark(aiob->hardware, 5);
	    } else {
		if (aiob->ignored) {
		    TPoint p = aiob->hardware->pos;
		    line(p+TPoint(-6,-6), p+TPoint(6,6), ClBlack);
		    line(p+TPoint(-6,6), p+TPoint(6,-6), ClBlack);
		}
	    }
    }
}

void AIGUI::showBuilding(AIObject* aiob, const RGB& c1, const RGB& c2, int layer) {
    HCC* hw = (HCC*)(aiob->hardware);
    if (layer == 1) {
	if (!hw->INVINCIBLE) {
	    int n = (hw->MAX_HP - hw->hp) * SMOKE_LENGTH / hw->MAX_HP;
	    if (n > 5) {
		int s = hw->SIZE / 3;
		showDamage(aiob, n);
		showDamage(aiob, n, TPoint(s, s));
		showDamage(aiob, n, TPoint(-s, s));
		showDamage(aiob, n, TPoint(s, -s));
		showDamage(aiob, n, TPoint(-s, -s));
	    }
	}
	rect(hw->x1, hw->y1, hw->x2, hw->y2,
	    (hw->hp == hw->old_hp[1]) ? c1 : c2);
    }
}

void AIGUI::showGraph(Graph* g, const RGB& nc, const RGB& ec) {
    if (!state.params.show_grid)
	return;
    for (int i=0; i<g->size; i++) {
	if (g->walls[i])
	    continue;
	for (int j=0; j<g->N_CNT; j++)
	    if (g->N_ARR[j] > 0) {
		RCHK(i+g->N_ARR[j], MAX_NODES);
		if (!g->walls[i+g->N_ARR[j]])
		    line(g->nodePos(i), g->nodePos(i+g->N_ARR[j]), ec);
	    }
    }
    for (int i=0; i<g->size; i++) {
	if (g->walls[i])
	    continue;
	circle(g->nodePos(i), 2, nc);
	if (state.params.show_blocked && (state.used_nodes[i]))
	    circle(g->nodePos(i), 2, ClMaroon);
	if (state.params.show_bfs_dist)
	    drawInt((int)(g->nodePos(i).x-3), (int)(g->nodePos(i).y-3), 3, 6, g->dist[i], ClNavy);
    }
}

void AIGUI::showBoundaries() {
    for (int i=0; i<state.boundc; i++) {
	Boundary* b = state.boundaries[i];
	line(b->x1, b->y1, b->x2, b->y2, ClBrown);
//	    ((state.game_time + i) % state.boundc == 0) ? ClOrange : ClBrown);
    }
    for (int i=0; i<state.MAP_WIDTH; i+=2)
	for (int j=0; j<state.MAP_HEIGHT; j+=2)
	    if (!state.accessible[i][j])
		putPixel(TPoint(i, j), ClBrown);
}

void AIGUI::drawGame1(int layer) {
    if (layer == 0) {
	showBoundaries();
	showGraph(state.map, ClGray, ClGray);
    } else if (layer == 1) {
        FOREACH(state.sheeps, it)
	    showUnit(it, ClGray, ClGray, layer);
	FOREACH(state.minerals, it)
	    showUnit(it, ClBlue, ClBlue, layer);
	FOREACH(state.workers, it)
	    showUnit(it, ClNavy, ClDarkGreen, layer);
	FOREACH(state.myControlCenters, it)
	    showBuilding(it, ClGreen, ClGreen, layer);
	drawInt(mapW-20, 10, 10, 18, state.total_resources, ClBlue);
	if (state.game_time < 40)
	    drawInt(40, 40, 10, 20, (int)state.minerals->
		nearest(NODEAT(state.myControlCenters->any()->
		hardware->pos)), ClWhite);
    }
}

void AIGUI::drawGame2(int layer) {
    if (layer == 0)
	showGraph(state.map, ClGray, ClGray);
    FOREACH(state.sheeps, it)
	showUnit(it, ClGray, ClGray, layer);
    FOREACH(state.myTanks, it)
	showUnit(it, ClNavy, ClBlue, layer);
    FOREACH(state.oppTanks, it)
	showUnit(it, ClMaroon, ClRed, layer);
    FOREACH(state.myControlCenters, it)
        showBuilding(it, ClGreen, ClBlue, layer);
    FOREACH(state.oppControlCenters, it)
        showBuilding(it, ClOrange, ClRed, layer);
    if (layer == 0)
	showBoundaries();
    if (layer == 1) {
	drawInt(50, 20, 25, 50, state.myTanks->size, ClBlue);
	drawInt(980, 20, 25, 50, state.oppTanks->size, ClRed);
	drawHp(state.myTanks, 100, 20, 930, 32, ClNavy, ClBlue);
	drawHp(state.oppTanks, 100, 35, 930, 47, ClMaroon, ClRed);
	if (state.params.show_profilers) {
	    showProfiler(0, 100, 50, 300, 70);
	    showProfiler(1, 730, 50, 930, 70);
	}
    };
}

void AIGUI::drawGame4(int layer) {
    if (layer == 0) {
	showBoundaries();
	showGraph(state.map, ClGray, ClGray);
    }
    FOREACH(state.sheeps, it)
	showUnit(it, ClGray, ClGray, layer);
    FOREACH(state.myMarines, it)
	showUnit(it, ClNavy, ClBlue, layer);
    FOREACH(state.oppMarines, it)
	showUnit(it, ClMaroon, ClRed, layer);
    if (layer == 1) {
	drawInt(50, 20, 25, 50, state.myMarines->size, ClBlue);
	drawInt(980, 20, 25, 50, state.oppMarines->size, ClRed);
	drawHp(state.myMarines, 100, 20, 930, 30, ClNavy, ClBlue);
	drawHp(state.oppMarines, 100, 32, 930, 42, ClMaroon, ClRed);
	if (state.params.show_profilers) {
	    showProfiler(0, 100, 50, 300, 70);
	    showProfiler(1, 730, 50, 930, 70);
	}
//	int n = state.myMarines->size - state.myMarines->outOfSync();
//	if (n == state.myMarines->size)
//	    n = 0;
//	drawInt(mapW/2-10, 50, 8, 16, -n, ClWhite);
    }
}
