#ifndef AIState_H
#define AIState_H

#include "GameStateModule.H"

#include "AIGUI.H"
#include "Boundary.H"
#include "Hardware.H"
#include "Graph.H"
#include "UnitGroup.H"
#include "AssaultFormation.H"
#include "WorkerFormation.H"

using namespace std;

struct AI_params {
    const char* f_path;
    int window_width;
    int window_height;
    int elen;		// edge length for current game, computed in MyEventHandler::detectGame()
    int elen1;
    int elen2;
    int elen3;
    int elen4;
    int sync_norm;
    bool show_unit_state;
    bool show_grid;
    bool show_blocked;
    bool show_dest;
    bool show_path;
    bool show_bfs_dist;
    bool show_profilers;
    bool show_start;
    bool show_end;
    bool show_values;
    bool show_prop;
    int speed;
    int look_ahead;
};

struct AIState_S {
    bool quit;
    bool initialized;		// main initialization completed
    bool aiinitialized;		// ai initialization completed
    int game_time;		// game ticks
    int player_id;
    int last_aiob_id;		// used in Hardware constructor
    int behind;			// delay (in frames (game ticks)) to(?) the server
    bool old_behind;		// 	in previous frame
    int handicap;		// speed handicap (%) -- client option
    int init_delay;		// initial delay in frames -- client option
    GameStateModule* gsm;
    AIGUI* gui;
    struct AI_params params;	// parameters values read from a text file
    int MAP_WIDTH;		// map size in pixels
    int MAP_HEIGHT;
    int SHEEP_SIZE;		// objects parameters: size, max hp, attack range and cooldown
    int MARINE_SIZE;
    int MAX_MARINE_HP;
    int MARINE_RANGE;
    int MARINE_COOLDOWN;
    int TANK_SIZE;
    int MAX_TANK_HP;
    int TANK_RANGE;
    int TANK_COOLDOWN;
    int CC_SIZE;
    int MAX_CC_HP;
    int MINERAL_SIZE;
    int WORKER_SIZE;
    int GAME_TYPE;		// game scenario (1 -- 4)
    int boundc;			// walls count
    Boundary* boundaries[MAX_BOUNDARIES];	// walls table
    bool accessible[MAX_WIDTH][MAX_HEIGHT];	// accessible points on map
    Graph* map;
    Hardware* used_nodes[MAX_NODES];		// graph nodes near game objects; blocked
    int total_resources;
    UnitGroup* allUnits;
    UnitGroup* sheeps;
    UnitGroup* myMarines;
    UnitGroup* oppMarines;
    UnitGroup* myTanks;
    UnitGroup* oppTanks;
    UnitGroup* myControlCenters;
    UnitGroup* oppControlCenters;
    UnitGroup* minerals;
    UnitGroup* workers;
    UnitGroup* myFighters;
    UnitGroup* enemyUnits;
    WorkerFormation* f0;
    AssaultFormation* f1;
};

extern AIState_S state;		// declared in MyEventHandler

#endif
