#include <iostream>
#include <sstream>
#include <string.h>

#include "Game.H"
#include "GameChanges.H"
#include "GameStateModule.H"
#include "Options.H"

#include "Common.H"
#include "AIState.H"
#include "Boundary.H"
#include "Hardware.H"
#include "UnitGroup.H"
#include "AIGUI.H"
#include "Coordinator1.H"
#include "Coordinator2.H"
#include "Coordinator4.H"
#include "MyEventHandler.H"

#define chk_gob() \
    gob = (*it)->get_GameObj();\
    if (!gob) continue;\
    if (!gob->sod.in_game) continue

#define check_val(name, pname) \
    if (key == name)\
	state.params.pname = val

#define check_bool(name, pname) \
    if (key == name)\
	state.params.pname = (val == 1)

AIState_S state;

using namespace std;

void MyEventHandler::read_params() {
    c_notice << "reading parameters..." << endl;
    state.params.window_width = 800;
    state.params.window_height = 600;
    state.params.elen = -1;
    state.params.elen1 = 10;
    state.params.elen2 = 10;
    state.params.elen3 = 10;
    state.params.elen4 = 10;
    state.params.sync_norm = 0;
    state.params.show_unit_state = false;
    state.params.show_grid = false;
    state.params.show_blocked = false;
    state.params.show_dest = false;
    state.params.show_path = false;
    state.params.show_bfs_dist = false;
    state.params.show_profilers = false;
    state.params.show_start = false;
    state.params.show_end = false;
    state.params.show_values = false;
    state.params.show_prop = false;
    state.params.speed = 10;
    state.params.look_ahead = 3;
    FILE* in = fopen(state.params.f_path, "r");
    if (in == NULL) {
	c_error << state.params.f_path << endl;
	ERR("params file not found!");
    }
    char buf[255]={0};
    while (fgets(buf, 255, in)) {
	if ((strlen(buf) < 3) || (buf[0] == '#'))
	    continue;
	string key;
	int val;
	istringstream is(buf);
	is >> key;
	is >> val;
	check_val("window_width", window_width);
	check_val("window_height", window_height);
	check_val("edge_length_1", elen1);
	check_val("edge_length_2", elen2);
	check_val("edge_length_3", elen3);
	check_val("edge_length_4", elen4);
	check_val("sync_norm", sync_norm);
	check_bool("show_unit_state", show_unit_state);
	check_bool("show_grid", show_grid);
	check_bool("show_blocked", show_blocked);
	check_bool("show_dest", show_dest);
	check_bool("show_path", show_path);
	check_bool("show_bfs_dist", show_bfs_dist);
	check_bool("show_profilers", show_profilers);
	check_bool("show_start", show_start);
	check_bool("show_end", show_end);
	check_bool("show_values", show_values);
	check_bool("show_prop", show_prop);
	check_val("speed", speed);
	check_val("look_ahead", look_ahead);
	c_notice << "  " << key << " = " << val << endl;
    }
    c_notice << "done" << endl;
}

inline bool border(TPoint p) {
    return ((p.x <= 0) || (p.y <= 0) || (p.x >= mapW) || (p.y >= mapH));
}

inline bool border(Boundary* b) {
    return border(b->p1()) || border(b->p2());
}

void initPixels() {
    int okc = 0;
    int nokc = 0;
    bool s_ok = true;
    for (int y=0; y<state.MAP_HEIGHT; y++) {
	for (int z=0; z<state.boundc; z++) {
	    Boundary* b = state.boundaries[z];
	    if (((b->x1 == 0) && (b->y1 == y))
		|| ((b->x2 == 0) && (b->y2 == y)))
		    s_ok = !s_ok;
	}
	bool ok = s_ok;
	for (int x=0; x<state.MAP_WIDTH; x++) {	
	    for (int z=0; z<state.boundc; z++) {
		Boundary* b = state.boundaries[z];
		if ((y<b->y1) || (y>=b->y2) || (b->y1==b->y2))
		    continue;
		int bx = b->x1 + (int)((b->x2-b->x1)*(1-((float)(b->y2-y)/(b->y2-b->y1))));
		if (x == bx) {
		    ok = !ok;
		}
		if (ok) {
		    okc++;
		} else {
		    nokc++;
		}
	    }
	    RCHK(x, MAX_WIDTH);
	    RCHK(y, MAX_HEIGHT);
	    state.accessible[x][y] = ok;
	}
    }
    if (okc < nokc)
	for (int x=0; x<state.MAP_WIDTH; x++)
	    for (int y=0; y<state.MAP_HEIGHT; y++)
		state.accessible[x][y] = !state.accessible[x][y];
}

void MyEventHandler::init_boundaries() {
    c_info << "computing pathing map..." << endl;
/*    if (state.GAME_TYPE == 1) {		// incomplete
	int bc = state.boundc;
	Boundary** b = state.boundaries;
	Boundary* b2[MAX_BOUNDARIES];
	int b2c = 0;
        bool used[MAX_BOUNDARIES];
	for (int i=0; i<bc; i++)
	    used[i] = false;
	for (int i=0; i<bc; i++)
	    if ((!used[i]) && (border(b[i]))) {
		used[i] = true;
		TPoint startp;
		TPoint endp;
		if (border(b[i]->p1())) {
		    startp = b[i]->p1();
		    endp = b[i]->p2();
		} else {
		    startp = b[i]->p2();
		    endp = b[i]->p1();
		}
		b2[b2c++] = b[i];
		int j = 0;
		while ((j != bc) && (startp != endp)) {
		    for (j=0; j<bc; j++) {
			if ((j == i) || (used[j]))
			    continue;
			if (b[j]->p1() == endp) {
			    b2[b2c++] = b[j];
			    endp = b[j]->p2();
			    used[j] = true;
			    break;
			} else if (b[j]->p2() == endp) {
			    b2[b2c++] = b[j];
//			    b[j]->reverse();
			    endp = b[j]->p1();
			    used[j] = true;
			    break;
			}
		    }
		}
	    }
	for (int i=0; i<bc; i++)
	    if (!used[i]) {
		used[i] = true;
		TPoint startp = b[i]->p1();
		TPoint endp = b[i]->p2();
		b2[b2c++] = b[i];
		while (endp != startp) {
		    for (int j=0; j<bc; j++) {
			if ((j == i) || (used[j]))
			    continue;
			if (b[j]->p1() == endp) {
			    b2[b2c++] = b[j];
			    endp = b[j]->p2();
			    used[j] = true;
			    break;
			}
			if (b[j]->p2() == endp) {
			    b2[b2c++] = b[j];
//			    b[j]->reverse();
			    endp = b[j]->p1();
			    used[j] = true;
			    break;
			}
		    }
		}
	    }
	state.boundc = bc = b2c;
	for (int i=0; i<bc; i++)
	    state.boundaries[i] = b[i] = b2[i];
	initPixels();
	b2c = 0;
	for (int i=0; i<bc-2; i++) {
	    if (border(b[i]) || border(b[i+1]) || border(b[i+2]))
		continue;
	    TPoint p = (b[i]->p1() + b[i+1]->p1() + b[i+2]->p1()) / 3;
	}
    }*/
    initPixels();
    state.map->addBoundaries();
    c_info << "done." << endl;
}

void MyEventHandler::beforeInitialization() {
    string par = "params.txt";
    Options::get("-param_path", par);
    state.params.f_path = par.c_str();
    read_params();
    state.initialized = false;
    state.aiinitialized = false;
    state.SHEEP_SIZE = -1;
    state.MARINE_SIZE = -1;
    state.TANK_SIZE = -1;
    state.CC_SIZE = -1;
    state.MINERAL_SIZE = -1;
    state.WORKER_SIZE = -1;
    state.GAME_TYPE = -1;
    state.allUnits = new UnitGroup();
    state.sheeps = new UnitGroup();
    state.myMarines = new UnitGroup();
    state.oppMarines = new UnitGroup();
    state.myTanks = new UnitGroup();
    state.oppTanks = new UnitGroup();
    state.myControlCenters = new UnitGroup();
    state.oppControlCenters = new UnitGroup();
    state.minerals = new UnitGroup();
    state.workers = new UnitGroup();
    state.myFighters = new UnitGroup();
    state.enemyUnits = new UnitGroup();
    state.gui = NULL;
    state.handicap = 0;
    state.init_delay = 0;
    state.f0 = NULL;
    state.f1 = NULL;
    int tilep = state.gsm->get_game().get_tile_points();
    state.MAP_WIDTH = state.gsm->get_game().get_map().get_width() * tilep;
    state.MAP_HEIGHT = state.gsm->get_game().get_map().get_height() * tilep;
    c_info << "map size: " << state.MAP_WIDTH << " x " << state.MAP_HEIGHT << endl;
    if ((state.MAP_WIDTH > MAX_WIDTH) || (state.MAP_HEIGHT > MAX_HEIGHT))
	ERR("map too large.");
    state.game_time = 0;
    state.player_id = state.gsm->get_game().get_client_player();
    state.last_aiob_id = 0;
    state.boundc = 0;
    state.total_resources = 0;
    c_info << "stage 1 initialization completed." << endl;
}

void MyEventHandler::aiinit() {
    if (state.aiinitialized)
	return;
    state.aiinitialized = true;
    c_notice << "initializing AI." << endl;
    if (state.GAME_TYPE == 1)
	aiInit1();
    if (state.GAME_TYPE == 2)
	aiInit2();
    if (state.GAME_TYPE == 4)
	aiInit4();
    c_notice << "AI initialized" << endl;
}

void MyEventHandler::detectGame() {
    c_info << "detecting game type..." << endl;
    if (state.workers->size > 0) {
	state.GAME_TYPE = 1;
	// TODO: game type 3 not recognized!
    } else {
	if (state.myTanks->size > 0) {
	    state.GAME_TYPE = 2;
	} else {
	    state.GAME_TYPE = 4;
	}
    }
    c_notice << "game scenario: " << state.GAME_TYPE << endl;
    if (state.GAME_TYPE == 1) {
	state.params.elen = state.params.elen1;
	state.map = new Graph();
    } else if (state.GAME_TYPE == 2) {
	state.params.elen = state.params.elen2;
	state.map = new Graph();
	FOREACH(state.myControlCenters, it)
	    it->hardware->INVINCIBLE = false;
	FOREACH(state.oppControlCenters, it)
	    it->hardware->INVINCIBLE = false;
    } else if (state.GAME_TYPE == 3) {
	state.params.elen = state.params.elen3;
	state.map = new Graph();
    } else if (state.GAME_TYPE == 4) {
	state.params.elen = state.params.elen4;
	state.map = new Graph();
        init_boundaries();
    } else {
	ERR("Unknown game scenario.");
    }
}

void MyEventHandler::afterInitialization() {
    state.initialized = true;
    if (state.init_delay > 0)
	c_notice << "initial delay set to " << state.init_delay << " frames" << endl;
    if (state.handicap > 0)
	c_notice << "handicap set to " << state.handicap << "%" << endl;
    if (state.GAME_TYPE == 1) {
	Hardware* cc = state.myControlCenters->any()->hardware;
	int x1 = ((HCC*)cc)->x1;
	int x2 = ((HCC*)cc)->x2;
	int y1 = ((HCC*)cc)->y1;
	int y2 = ((HCC*)cc)->y2;
	for (int x=x1; x<=x2; x++)
	    for (int y=y1; y<=y2; y++)
		state.accessible[x][y] = false;
    }
    if (state.gui)
	state.gui->initGui();
    c_info << "stage 2 initialization completed." << endl;
}

void MyEventHandler::removeObj(GameObj* gob) {
    int own = *gob->sod.owner;
    if (own == 2) {
	state.sheeps->removeUnit(gob);
    } else {
	if (own == state.player_id) {
	    state.myMarines->removeUnit(gob);
	} else {
	    state.oppMarines->removeUnit(gob);
	}
    }
    state.allUnits->removeUnit(gob);
}

void MyEventHandler::updateWorld() {
    new_boundaries = false;
    state.game_time = state.gsm->get_game().tick();
    const GameChanges* changes = &state.gsm->get_changes();
    GameObj* gob;
    FORALL(changes->new_objs, it) {
	chk_gob();
	int own = *gob->sod.owner;
	string name = gob->bp_name();
	if ((state.SHEEP_SIZE == -1) && (gob->bp_name() == "sheep")) {
	    state.SHEEP_SIZE = *gob->sod.radius;
	    c_notice << "sheep size = " << state.SHEEP_SIZE << endl;
	} else if ((state.MINERAL_SIZE == -1) && (name == "mineral")) {
	    state.MINERAL_SIZE = *gob->sod.radius;
	    c_notice << "mineral size = " << state.MINERAL_SIZE << endl;
	}
	if (own != state.player_id)
	    continue;
	if ((state.MARINE_SIZE == -1) && (name == "marine")) {
	    state.MARINE_SIZE = *gob->sod.radius;
	    state.MAX_MARINE_HP = gob->get_int("hp");
	    state.MARINE_RANGE = gob->component("weapon")->get_int("max_ground_range");
	    state.MARINE_COOLDOWN = gob->component("weapon")->get_int("cooldown");// FIXME! czy bedzie tutaj dostepny?
	    c_notice << "marine size = " << state.MARINE_SIZE << endl;
	    c_notice << "max marine hp = " << state.MAX_MARINE_HP << endl;
	    c_notice << "marine range = " << state.MARINE_RANGE << endl;
	    c_notice << "marine cooldown = " << state.MARINE_COOLDOWN << endl;
	} else if ((state.TANK_SIZE == -1) && (name == "tank")) {
	    state.TANK_SIZE = *gob->sod.radius;
	    state.MAX_TANK_HP = gob->get_int("hp");
	    state.TANK_RANGE = gob->component("weapon")->get_int("max_ground_range");
	    state.TANK_COOLDOWN = gob->component("weapon")->get_int("cooldown");//FIXME!
	    c_notice << "tank size = " << state.TANK_SIZE << endl;
	    c_notice << "max tank hp = " << state.MAX_TANK_HP << endl;
	    c_notice << "tank range = " << state.TANK_RANGE << endl;
	    c_notice << "tank cooldown = " << state.TANK_COOLDOWN << endl;
	} else if ((state.CC_SIZE == -1) && (name == "controlCenter")) {
	    state.CC_SIZE = *gob->sod.radius;
	    state.MAX_CC_HP = gob->get_int("hp");
	    c_notice << "control center size = " << state.CC_SIZE << endl;
	    c_notice << "control center hp = " << state.MAX_CC_HP << endl;
	} else if ((state.WORKER_SIZE == -1) && (name == "worker")) {
	    state.WORKER_SIZE = *gob->sod.radius;
	    c_notice << "worker size = " << state.WORKER_SIZE << endl;
	}
    }
    FORALL(changes->new_objs, it) {
	chk_gob();
	int own = *gob->sod.owner;
	string name = gob->bp_name();
	if (name == "start_loc")
	    continue;
	AIObject* aiob = new AIObject(gob);
	if (name == "marine") {
	    if (own == state.player_id) {
		state.myFighters->addUnit(aiob);
		state.myMarines->addUnit(aiob);
	    } else {
		state.oppMarines->addUnit(aiob);
		state.enemyUnits->addUnit(aiob);
	    }
	} else if (name == "tank") {
	    if (own == state.player_id) {
		state.myFighters->addUnit(aiob);
		state.myTanks->addUnit(aiob);
	    } else {
		state.oppTanks->addUnit(aiob);
		state.enemyUnits->addUnit(aiob);
	    }
	} else if (name == "controlCenter") {
	    if (own == state.player_id) {
		state.myControlCenters->addUnit(aiob);
	    } else {
		state.oppControlCenters->addUnit(aiob);
		state.enemyUnits->addUnit(aiob);
	    }
	} else if (name == "sheep") {
	    state.sheeps->addUnit(aiob);
	} else if (name == "mineral") {
	    state.minerals->addUnit(aiob);
	} else if (name == "worker") {
	    state.workers->addUnit(aiob);
	} else {
	    c_warning << "Unidentified game object: " << gob->bp_name() << endl;
	}
	state.allUnits->addUnit(aiob);
    }
    if (!state.initialized)
	c_info << "game objects initialized" << endl;
    if (state.GAME_TYPE == -1)
	detectGame();
    FORALL(changes->new_boundaries, it) {
	gob = (*it)->get_GameObj();
	if (gob == NULL)
	    continue;
	new_boundaries = true;
	state.boundaries[state.boundc] = new Boundary(
	    *gob->sod.x1, *gob->sod.y1, *gob->sod.x2, *gob->sod.y2);
	state.boundc++;
	c_debug << "added boundary(" << *gob->sod.x1 << ", " << *gob->sod.y1;
	c_debug << ", " << *gob->sod.x2 << ", " << *gob->sod.y2 << ")" << endl;
    }
    FORALL(changes->vanished_objs, it) {
	chk_gob();
	removeObj(gob);
    }
    if (new_boundaries)
	init_boundaries();
    if (state.behind > 0)
	return;
    for (int i=0; i<MAX_NODES; i++)
	state.used_nodes[i] = NULL;
    FOREACH(state.allUnits, it)
	it->update();
    state.allUnits->refresh();
    state.sheeps->refresh();
    state.myMarines->refresh();
    state.oppMarines->refresh();
    state.myTanks->refresh();
    state.oppTanks->refresh();
    state.myControlCenters->refresh();
    state.oppControlCenters->refresh();
    state.minerals->refresh();
    state.workers->refresh();
    state.myFighters->refresh();
    state.enemyUnits->refresh();
    return; //
    FOREACH(state.allUnits, it) {
	int npos = state.map->nodeAt(it->hardware->pos);
	RCHK(npos, MAX_NODES);
	state.used_nodes[npos] = it->hardware;
	if (it->owner != state.player_id)
	    for (int d=0; d<state.map->N_CNT; d++) {
		int npos = state.map->nodeAt(it->hardware->pos) + state.map->N_ARR[d];
		RCHK(npos, MAX_NODES);
		state.used_nodes[npos] = it->hardware;
	    }
    }
}

bool MyEventHandler::handle_event(const Event& e) {
    if (e.get_who() != GameStateModule::FROM)
	return false;
    if (e.get_what() == GameStateModule::STOP_MSG) {
	state.quit = true;
	return true;
    }
    if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
	state.quit = true;
	return true;
    }
    if (e.get_what() == GameStateModule::VIEW_MSG) {
	if (state.initialized)
	    AI_GUI->timeStart(2, ClGray);
	state.old_behind = state.behind > 0;
	updateWorld();
	Game* game = &state.gsm->get_game();
	state.behind = game->get_view_frame() - game->get_action_frame();
	if (!state.initialized)
	    afterInitialization();
	if ((state.enemyUnits->size == 0) && (state.GAME_TYPE != 1)) {
	    state.quit = true;
	    return true;
	}
	if (state.gui) {
	    state.gui->event();
	    state.gui->beforeDisplay();
	    AI_GUI->timeStop(2);
	    state.gui->display();
	}
	AI_GUI->clearTime(2);
	AI_GUI->timeStart(2, ClGray);
	if (state.behind > 0) {
	    c_info << "BEHIND ";
	    if (state.aiinitialized) {
		c_info << MIN(state.behind, state.game_time - state.behind + 1);
	    } else {
		c_info << state.behind;
	    }
	    c_info << " FRAMES." << endl;
	} else {
    	    char* line = "--------------------------------------------";
	    c_info << line << " frame " << state.game_time << " " << line << endl;
	    if ((state.game_time > state.init_delay)
		&& ((state.myFighters->size > 0) || (state.GAME_TYPE % 2 == 1))) {
		aiinit();
		AI_GUI->timeStart(2, ClNavy);
		if (state.GAME_TYPE == 1)
		    computeActions1();
		if (state.GAME_TYPE == 2)
		    computeActions2();
		if (state.GAME_TYPE == 4)
		    computeActions4();
		AI_GUI->timeStart(2, ClGray);
	    }
	}
	state.gsm->send_actions();
	if (state.gui) {
	    state.gui->afterDisplay();
	    if (state.gui->quit) {
		cout << "exiting..." << endl;
		exit(0);
	    }
	}
	AI_GUI->timeStart(2, ClBrown);
	return true;
    }
    return false;
}
