#ifndef AIGUI_H
#define AIGUI_H

#include "GUI.H"

#include "TPoint.H"
#include "Graph.H"
#include "UnitGroup.H"

#define MAXPROFILERS	3
#define MAXPROFILETIME	1250
#define TIMEREDLINE	1000

#define ClBackground	RGB( 64,  64,  96)
#define ClBlack		RGB(  0,   0,   0)
#define ClBlue		RGB(  0,   0, 255)
#define ClBrown		RGB( 80,  40,   0)
#define ClDarkGreen	RGB(  0, 128,   0)
#define ClDarkRed	RGB( 64,   0,   0)
#define ClGray		RGB(128, 128, 128)
#define ClGreen		RGB(  0, 255,   0)
#define ClMaroon	RGB(128,   0,   0)
#define ClNavy		RGB(  0,   0, 128)
#define ClOrange	RGB(255, 128,   0)
#define ClRed		RGB(255,   0,   0)
#define ClYellow	RGB(255, 255,   0)
#define ClWhite		RGB(255, 255, 255)

using namespace std;

class Graph;

class AIGUI : public GUI {
    private:
	int s_arr[MAX_UNITS][SMOKE_LENGTH][3];
	int starts[MAXPROFILERS];
	int times[MAXPROFILERS];
	RGB colors[MAXPROFILERS];
	RGB msecs[MAXPROFILERS][MAXPROFILETIME];
	int currTime();
	void clear_screen();
	void drawDigth(int x, int y, int w, int h, int n, const RGB& c);
	void drawHp(UnitGroup* units, int x1, int y1, int x2, int y2, const RGB& fc, const RGB& sc);
	void showDamage(AIObject* aiob, int n, TPoint dpos = TPoint(0, 0));
	void showUnit(AIObject* aiob, const RGB& c1, const RGB& c2, int layer);
	void showBuilding(AIObject* aiob, const RGB& c1, const RGB& c2, int layer);
	void showGraph(Graph* g, const RGB& nc, const RGB& ec);
	void showBoundaries();
	void showProfiler(int id, int x1, int y1, int x2, int y2);
	void drawGame1(int layer);
	void drawGame2(int layer);
	void drawGame4(int layer);
    public:
        AIGUI();
	~AIGUI() {};
	void initGui();		// after initialization
	void beforeDisplay();	// start drawing current frame
        void display();		// draw standard gui
	void afterDisplay();	// end drawing current frame
	void line(int x1, int y1, int x2, int y2, const RGB& c);
	void line(TPoint p1, TPoint p2, const RGB& c);
	void line(int node1, int node2, const RGB& c);
	void circle(int x, int y, int r, const RGB& c, bool filled = false);
	void circle(TPoint s, int r, const RGB& c, bool filled = false);
	void circle(int node, int r, const RGB& c, bool filled = false);
	void putPixel(TPoint p, const RGB& c);
	void rect(int x1, int y1, int x2, int y2, const RGB& c);
	void filledRect(int x1, int y1, int x2, int y2, const RGB& c);
	void mark(Hardware* h, int type);
	void drawInt(int x, int y, int w0, int h, int n, const RGB& c); // n -- int value
	void clearTime(int id);
	void timeStart(int id, const RGB& color);
	int timeStop(int id);
};

#endif
