#include "AIState.H"
#include "UnitGroup.H"
#include "Formation.H"
#include "AssaultFormation.H"

AssaultFormation::AssaultFormation() : Formation() {};

bool AssaultFormation::ignore(AIObject* aiob) {
    ab_init = false;
    if (aiob->hardware->HW_TYPE == HW_CC)
	return (state.oppTanks->size > 5) && (state.game_time < 5000);
    int allyc = 0;
    FOREACH(state.enemyUnits, opp)
	if (opp->hardware->CAN_SHOOT)
	    if (dist(aiob->hardware->pos, opp->hardware->pos) < 60)
		allyc++;
    return allyc < 3;
}

bool AssaultFormation::startFrom(int node) {
    RCHK(node, MAX_NODES);
    float d = state.enemyUnits->nearest(node);
    float r = any()->hardware->RANGE + 2*any()->hardware->SIZE;
    int w = state.params.elen;
    if (d <= r-w)
	is_blocked[node] = true;
    if ((d > r-w) && (d < r)) {
	return canSee(NODEPOS(node), state.enemyUnits->nearest(NODEPOS(node))->hardware->pos);
    } else {
	return false;
    }
}

float AssaultFormation::evalStart(int node) {
    float val = 0;
    float d0 = state.enemyUnits->nearest(node);
    for (int i=0; i<10; i++) {
	float d = state.enemyUnits->nearest(node, i);
	val += (d - d0) / (i+2);
    }
    return val;
}

bool in_range(AIObject* a1, TPoint p) {
    return dist(a1->hardware->pos, p)
	- 2 * a1->hardware->SIZE < a1->hardware->RANGE;
}

float AssaultFormation::evalEnd(int node) {
    float d = state.enemyUnits->nearest(node);
    if (state.myFighters->size - state.enemyUnits->size > 15)
	return 100/d;
    if (state.oppMarines->size + state.oppTanks->size < 3)
	return 100/d;
    float r = ABS(d - avg_dist + state.params.speed + state.game_time/1000);
    float v = 0;
    FOREACH(this, it) {
	float d = dist(it->hardware->pos, NODEPOS(node));
	v += 10 / d;
    }
    if (!ab_init) {
	FOREACH(state.allUnits, it)
	    if (it->owner < 2) {
		int n = 0;
		FOREACH(state.allUnits, opp)
		    if (it->owner == 1 - opp->owner)
			if (in_range(it, opp->hardware->pos))
			    n++;
		RCHK(it->hardware->id, MAX_UNITS);
		attacked_by[it->hardware->id] = n;
	    }
	ab_init = true;
    }
    float x = 0;
    FOREACH(state.enemyUnits, it)
	if (in_range(it, NODEPOS(node)))
	    x -= 1 / (attacked_by[it->hardware->id] + 1);
    if (x > 0)
	x += 1;
    return 10*x - r;
}
