#include <iostream>

#include "Game.H"
#include "GameStateModule.H"
#include "Options.H"
#include "SDLinit.H"
#include "ShutDown.H"

#include "Common.H"
#include "AIState.H"
#include "MyEventHandler.H"

static char* VERSION = "ai client v1.0";

using namespace std;

void add_options() {
    Options opt("AI client");
    opt.put("-go_disp", "2d debug display");
    opt.put("-param_path", string("params.txt"), "specify path to parameters file");
    opt.put("-handicap", 0, "units speed handicap (%)");
    opt.put("-init_delay", 0, "initial delay (frames)");
}

int main(int argc, char** argv) {
    signals_shut_down(true);
    GameStateModule::Options::add();
    Game::Options::add();
    GUI::add_options();
    add_options();
    if (Options::process(argc, argv, c_error, VERSION))
	exit(20);
    SDLinit::network_init();
    GameStateModule::Options gsmo;
    GameStateModule gsm(gsmo);
    state.gsm = &gsm;
    MyEventHandler eh;
    gsm.add_handler(&eh);
    if (!gsm.connect())
	ERR("connection problems");
    eh.beforeInitialization();
    bool go_disp;
    Options::get("-go_disp", go_disp);
    if (go_disp)
	state.gui = new AIGUI();
    Options::get("-handicap", state.handicap);
    Options::get("-init_delay", state.init_delay);
    while (!state.quit)
	if (!gsm.recv_view())
	    SDL_Delay(1);
    c_notice << "game over." << endl;
    if (state.GAME_TYPE == 4)
	if (state.enemyUnits->size > 0) {
	    c_notice << (go_disp ? "red" : "blue") << " won (";
	    c_notice << state.oppMarines->size << " units left)." << endl;
	} else {
	    c_notice << (go_disp ? "blue" : "red") << " won (";
	    c_notice << state.myMarines->size << " units left)." << endl;
	}
    if (state.GAME_TYPE == 2)
	if (state.oppControlCenters->size > 0) {
	    c_notice << (go_disp ? "red" : "blue") << " won (";
	    c_notice << state.oppControlCenters->size << " buildings left)." << endl;
	} else {
	    c_notice << (go_disp ? "blue" : "red") << " won (";
	    c_notice << state.myControlCenters->size << " buildings left)." << endl;
	}
    if (state.GAME_TYPE == 1)
	c_notice << "gathered " << state.total_resources << " minerals." << endl;
    return 0;
}
