#include <vector>

#include "Common.H"
#include "AIState.H"
#include "Graph.H"
#include "AIGUI.H"
#include "UnitGroup.H"
#include "Formation.H"

Formation::Formation() : UnitGroup() {}

void Formation::sync() {
    FOREACH(this, it)
	it->software->sync();
}

bool Formation::blocked(int node) {
    return is_blocked[node] || state.map->walls[node] || state.used_nodes[node];
}

bool Formation::startFrom(int node) {
    return false;
}

bool Formation::endAt(int node) {
    FOREACH(this, it)
	for (int d=0; d<state.map->N_CNT; d++)
	    if (it->software->curr_node == node + state.map->N_ARR[d])
		return true;
    return false;
}

float Formation::evalStart(int node) {
    return 1;
}

float Formation::evalNext(int node) {
    float val = 0;
    int cnt = 0;
    for (int i=0; i<NCNT; i++) {
	if (nb[i]) {
	    int n = node+NARR[i];
	    val += values[n];
	    cnt++;
	}
    }
    ASSERT(cnt > 0, "no neighbour found");
    return val / cnt;
}

float Formation::evalEnd(int node) {
    return 0;
}

void Formation::setDirections() {
    int used[MAX_NODES];
    for (int i=0; i<MAX_NODES; i++)
	used[i] = blocked(i);
    for (int i=0; i<size; i++) {
	AIObject* aiob = v[i];
	if (aiob->software->curr_node == -1) {
	    aiob->software->sync();
	    continue;
	}
	if ((!aiob->software->ready) && (aiob->software->time_stopped < 8))
	    continue;
	RCHK(aiob->software->curr_node, MAX_NODES);
	float max_val = (is_start[aiob->software->curr_node]) ? MAXINT : -MAXINT;
	for (int d=0; d<state.map->N_CNT; d++) {
	    int n = aiob->software->curr_node + state.map->N_ARR[d];
	    RCHK(n, MAX_NODES);
	    if (used[n])
		continue;
	    if (aiob->software->prev_node == n)
		continue;
	    if (values[n] > max_val) {
		max_val = values[n];
		next_dest[i] = n;
		used[n] = true;
	    }
	}
    }
}

void Formation::redraw() {
    int map_size = state.map->size;
    if (state.params.show_blocked)
	for (int i=0; i<map_size; i++)
	    if (is_blocked[i]) {
		DRAWLINE(NODEPOS(i) + TPoint(-3, -3), NODEPOS(i) + TPoint(3, 3), ClGray);
		DRAWLINE(NODEPOS(i) + TPoint(-3, 3), NODEPOS(i) + TPoint(3, -3), ClGray);
	    }
    if (state.params.show_start)
	for (int i=0; i<map_size; i++)
	    if (is_start[i])
		DRAWCIRCLE(i, 3, ClBlack);
    if (state.params.show_end)
	for (int i=0; i<map_size; i++)
	    if (is_end[i])
		DRAWCIRCLE(i, 3, ClWhite);
    if (state.params.show_values) {
	int max = 1;
	for (int i=0; i<map_size; i++)
	    if (values[i] > max)
		max *= 2;
	for (int i=0; i<map_size; i++)
	    if (values[i] > 0)
		DRAWCIRCLE(i, 2, RGB(255, MIN((int)(values[i] / max * 255), 255), 0), true);
    }
    if (state.params.show_bfs_dist)
	AI_GUI->drawInt(mapW/2, 60, 8, 16, (int)avg_dist, ClYellow);
}

bool Formation::ignore(AIObject* aiob) {
    return false;
}

void Formation::commit() {
    bool updating = true;
    int maxoos = state.params.sync_norm;
    maxoos = MAX(0, MIN(maxoos, state.myFighters->size-maxoos));
    if (outOfSync() <= maxoos) {
	c_info << "updating orders ..." << endl;
	AI_GUI->clearTime(1);
	AI_GUI->timeStart(1, ClGray);
	int map_size = state.map->size;
	RCHK(map_size, MAX_NODES);
	for (int i=0; i<map_size; i++) {
	    is_blocked[i] = is_start[i] = is_end[i] = false;
	    state.map->blocked[i] = state.used_nodes[i];
	}
	int ignc = 0;
	FOREACH(state.enemyUnits, opp)
	    if (opp->ignored = ignore(opp))
		ignc++;
	if (ignc >= state.enemyUnits->size / 2)
	    FOREACH(state.enemyUnits, opp)
		opp->ignored = false;
	avg_dist = 0;
	if (state.enemyUnits->size > 0)
	    avg_dist = distTo(state.enemyUnits);
	vector<int> start;
	AI_GUI->timeStart(1, ClDarkRed);
	for (int i=0; i<map_size; i++) {
	    values[i] = -MAXINT;
	    if (state.map->walls[i])
		continue;
	    if (state.used_nodes[i])
		state.map->blocked[i] = true;
	    if (startFrom(i)) {
		is_start[i] = true;
		start.push_back(i);
		values[i] = evalStart(i);
	    }
	    state.map->blocked[i] &= is_blocked[i];
	}
	int maxd = 0;
	AI_GUI->timeStart(1, ClWhite);
	state.map->BFS(start, -1);
	AI_GUI->timeStart(1, ClGray);
	for (int i=0; i<MAX_MAP_DIST; i++)
	    nodesInC[i] = 0;
	for (int i=0; i<map_size; i++) {
	    int d = state.map->dist[i];
	    if ((d < 0) || (state.map->walls[i]) || (state.map->boundary[i]))
		continue;
	    RCHK(d, MAX_MAP_DIST);
	    RCHK(nodesInC[d], MAX_MAP_DIST);
	    nodesInA[d][nodesInC[d]] = i;
	    nodesInC[d]++;
	    if (d > maxd)
		maxd = d;
	}
	AI_GUI->timeStart(1, ClYellow);
	for (int d=1; d<=maxd; d++)
	    for (int i=0; i<nodesInC[d]; i++) {
		RCHK(i, MAX_MAP_DIST);
		RCHK(i, MAX_NODES);
		int o = nodesInA[d][i];
		RCHK(o, MAX_NODES);
		for (int j=0; j<state.map->N_CNT; j++) {
		    int n = o + state.map->N_ARR[j];
		    RCHK(n, MAX_NODES);
		    nb[j] = false;
		    if (state.map->dist[n] == d-1) {
			if (state.params.show_prop)
			    DRAWLINE(o, n, ClBrown);
			nb[j] = true;
		    }
		}
		values[o] = evalNext(o);
	    }
	AI_GUI->timeStart(1, ClNavy);
	nc = 0;
	for (int i=0; i<map_size; i++)
	    if ((values[i] >= 0) && (endAt(i))) {
		is_end[i] = true;
		values[i] += evalEnd(i);
		RCHK(nc, MAX_NODES);
		next_ok[nc] = i;
		nc++;
	    }
	for (int i=0; i<size; i++)
	    next_dest[i] = -1;
	AI_GUI->timeStart(1, ClBlue);
	setDirections();
	AI_GUI->timeStart(1, ClDarkGreen);
	for (int i=0; i<size; i++)
	    if (next_dest[i] >= 0)
		v[i]->software->move(next_dest[i]);
	AI_GUI->timeStart(1, ClGray);
    } else {
	AI_GUI->clearTime(0);
	AI_GUI->timeStart(0, ClGray);
	updating = false;
    }
    redraw();
    FOREACH(this, it)
	it->software->commit();
    if (updating) {
	AI_GUI->timeStop(1);
    } else {
	AI_GUI->timeStop(0);
    }
}
