#include <vector>

#include "Common.H"
#include "TPoint.H"
#include "Boundary.H"
#include "AIState.H"
#include "AIGUI.H"
#include "GraphShape.H"
#include "Graph.H"

using namespace std;

Graph::Graph() {
    for (int i=0; i<MAX_NODES; i++) {
	walls[i] = blocked[i] = used[i] = true;
	dist[i] = -1;
    }
    for (int i=0; i<size; i++) {
	 boundary[i] = walls[i] = blocked[i] = (nodePos(i).x < 0)
	    || (nodePos(i).x > state.MAP_WIDTH) || (nodePos(i).y < 0)
	    || (nodePos(i).y > state.MAP_HEIGHT);
    }
    clearBFS();
}

TPoint Graph::round(TPoint p) {
    return nodePos(nodeAt(p));
}

void Graph::clearBFS() {
    for (int i=0; i<size; i++) {
	dist[i] = -1;
	used[i] = blocked[i] = walls[i];
    }
}

void Graph::addBoundaries() {
    for (int i=0; i<size; i++)
	walls[i] = boundary[i];
    for (int i=0; i<size; i++) {
	TPoint p = nodePos(i);
	for (int dx=-1; dx<2; dx++)
	    for (int dy=-1; dy<2; dy++)
		if (!state.accessible[(int)(p.x)+dx*MIN_BDIST][(int)(p.y)+dy*MIN_BDIST])
		    walls[i] = true;
    }
    int max = MAX_ISLAND_SIZE;
    while (BFS(rand() % size, -1) < max--);
    for (int i=0; i<size; i++)
        if (dist[i] == -1)
	    walls[i] = true;
    clearBFS();
}

int Graph::BFS(vector<int> start, int maxd, bool reversed) {
    int kolp = 0;
    int kolk = start.size();
    int time = 0;
    int last = kolk-1;
    RCHK((int)(start.size()), MAX_NODES);
    for (int i=0; i<(int)(start.size()); i++) {
	kol[i] = start[i];
	dist[i] = time;
    }
    for (int i=0; i<size; i++) {
	used[i] = blocked[i] || walls[i];
	if (reversed)
	    used[i] = !used[i];
	dist[i] = -1;
    }
    while (kolp < kolk) {
	int v = kol[kolp];
	RCHK(v, MAX_NODES);
	if (!used[v]) {
	    dist[v] = time;
	    used[v] = true;
	    for (int i=0; i<N_CNT; i++) {
		RCHK(kolk, MAX_NODES);
		kol[kolk] = v+N_ARR[i];
		kolk++;
	    }
	}
	kolp++;
	if (kolp > last) {
	    last = kolk-1;
	    if (time == maxd)
		break;
	    time++;
	}
    }
    return time;
}

int Graph::BFS(int start, int maxd) {
    vector<int> v;
    v.push_back(start);
    return BFS(v, maxd);
}

int Graph::nearbyOk(int node) {
    while ((blocked[node]) || (walls[node])) {
	int new_node = node + N_ARR[rand() % N_CNT];
	if ((new_node < 0) || (new_node >= size))
	    continue;
	if (boundary[new_node])
	    continue;
	node = new_node;
    }
    return node;
}
