#include "TPoint.H"
#include "AIState.H"
#include "GraphShape.H"

using namespace std;

GraphShape::GraphShape() {
    width = state.MAP_WIDTH / state.params.elen + 2;
    height = state.MAP_HEIGHT / (state.params.elen - 2) + 2;
    size = width * height - height / 2;
    ASSERT(size < MAX_NODES, "graph size");
    N_CNT = 6;
    N_ARR[0] = -width+1;
    N_ARR[1] = -width;
    N_ARR[2] = -1;
    N_ARR[3] = 1;
    N_ARR[4] = width-1;
    N_ARR[5] = width;
    c_info << width << " x " << height << endl;
    c_info << "graph size: " << size << endl;
}

int GraphShape::nodeAt(TPoint p) {
    int n;
    int ex = state.params.elen;
    int ey = ex - 2;
    int y = (int)((p.y + 3 * ey / 2) / ey);
    if (y % 2 == 1)
	p.x -= ex / 2;
    int x = (int)((p.x + 3 * ex / 2) / ex);
    if (y % 2 == 0) {
	n = (y/2)*(2*width-1)+x;
    } else {
	n = (y/2)*(2*width-1)+width+x;
    }
    if (n < 0)
	c_error << p.toString() << " --> " << n << endl;
    return n;
}

TPoint GraphShape::nodePos(int node) {
    if (node < 0)
	return nowhere;
    int ex = state.params.elen;
    int ey = ex - 2;
    int y = node / (2*width-1)-1;
    int x = node % (2*width-1);
    if (x < width) {
	y = 2*y+1;
    } else {
	y = 2*y+2;
	x -= width;
    }
    x--;
    TPoint p = TPoint(x * ex + (y % 2 ? 0 : ex / 2), y * ey);
    if (nodeAt(p) != node)
	c_info << node << " --> " << p.toString() << " --> " << nodeAt(p) << endl;
    ASSERT(nodeAt(p) == node, "graph shape");
    return p;
}
