#include "GameObj.H"

#include "Common.H"
#include "TPoint.H"
#include "Boundary.H"
#include "AIState.H"
#include "Hardware.H"

using namespace std;

Hardware::Hardware(GameObj* gob) {
    this->gob = gob;
    dead = false;
    MAX_HP = -1;
    MAX_COOLDOWN = -1;
    SIZE = -1;
    RANGE = -1;
    CAN_MOVE = true;
    CAN_SHOOT = false;
    INVINCIBLE = false;
    HW_TYPE = HW_UNKNOWN;
    pos = nowhere;
    hp = -1;
    cooldown = 0;
    id = state.last_aiob_id++;
    target = NULL;
    ally = (*gob->sod.owner == state.player_id);
}

void Hardware::afterInit() {
    update();
    for (int i=0; i<HP_HISTORY_LENGTH; i++)
	old_hp[i] = hp;
}

float Hardware::distTo(Hardware* h) {
    return gob->distance_to(*h->gob);
}

bool Hardware::isMoving() {
    return *gob->sod.speed > 0;
}

Hardware* Hardware::detectAttacker(float e) {
    Hardware* attacker = NULL;
    int dd = gob->dir_dmg;
    for (int i=0; i<32; i++) {
	if ((dd & (1 << i)) == 0)
	    continue;
	FOREACH(state.allUnits, opp) {
	    if (opp->owner != 1 - state.player_id)
		continue;
	    if (abs(angle(pos, opp->hardware->pos) - i) > e)
		continue;
	    if (distTo(opp->hardware) > opp->hardware->RANGE)
		continue;
	    if (opp->hardware->cooldown > 0)
		continue;
	    opp->hardware->cooldown = opp->hardware->MAX_COOLDOWN;
	    opp->hardware->target = this;
	    attacker = opp->hardware;
	}
    }
    return attacker;
}

void Hardware::update() {
    pos.x = *gob->sod.x;
    pos.y = *gob->sod.y;
    if (!INVINCIBLE)
	hp = gob->get_int("hp");
    if (ally) {
	cooldown = MAX_COOLDOWN-(state.game_time-gob->component("weapon")->get_int("last_fired"));
    } else if (cooldown > 0) {
	cooldown--;
    }
    if (target)
	if (target->dead)
	    target = NULL;
    if ((cooldown < 0) || (!target))
	cooldown = 0;
    for (int i=HP_HISTORY_LENGTH-1; i>0; i--)
	old_hp[i] = old_hp[i-1];
    old_hp[0] = hp;
    if (ally && (hp < old_hp[1])) {
        detectAttacker(0.5);
        detectAttacker(1);
        detectAttacker(1.5);
	float d = MAXINT;
	FOREACH(state.enemyUnits, it)
	    if (dist(it->hardware->pos, pos) < d)
		d = dist(it->hardware->pos, pos);
    }
}

HSheep::HSheep(GameObj* gob) : Hardware(gob) {
    SIZE = state.SHEEP_SIZE;
    INVINCIBLE = true;
    HW_TYPE = HW_SHEEP;
    afterInit();
}

HMarine::HMarine(GameObj* gob) : Hardware(gob) {
    MAX_HP = state.MAX_MARINE_HP;
    MAX_COOLDOWN = state.MARINE_COOLDOWN;
    SIZE = state.MARINE_SIZE;
    RANGE = state.MARINE_RANGE;
    CAN_SHOOT = true;
    HW_TYPE = HW_MARINE;
    afterInit();
}

HTank::HTank(GameObj* gob) : Hardware(gob) {
    MAX_HP = state.MAX_TANK_HP;
    MAX_COOLDOWN = state.TANK_COOLDOWN;
    SIZE = state.TANK_SIZE;
    RANGE = state.TANK_RANGE;
    CAN_SHOOT = true;
    HW_TYPE = HW_TANK;
    afterInit();
}

HCC::HCC(GameObj* gob) : Hardware(gob) {
    MAX_HP = state.MAX_CC_HP;
    hp = MAX_HP;
    SIZE = state.CC_SIZE;
    CAN_MOVE = false;
    HW_TYPE = HW_CC;
    x1 = MIN(*gob->sod.x1, *gob->sod.x2);
    y1 = MIN(*gob->sod.y1, *gob->sod.y2);
    x2 = MAX(*gob->sod.x1, *gob->sod.x2);
    y2 = MAX(*gob->sod.y1, *gob->sod.y2);
    pos.x = (x1+x2)/2;
    pos.y = (y1+y2)/2;
    RCHK(state.boundc, MAX_BOUNDARIES-4);
    afterInit();
}

void HCC::update() {
    hp = gob->get_int("hp");
    for (int i=HP_HISTORY_LENGTH-1; i>0; i--)
	old_hp[i] = old_hp[i-1];
    old_hp[0] = hp;
    if (ally && (hp < old_hp[1])) {
        detectAttacker(0.5);
        detectAttacker(1);
        detectAttacker(1.5);
    }
    if (state.map != NULL)
	for (int i=-SIZE; i<=SIZE; i++)
	    for (int j=-SIZE; j<=SIZE; j++) {
		RCHK(NODEAT(TPoint(i, j) + pos), MAX_NODES);
		state.used_nodes[NODEAT(TPoint(i, j) + pos)] = this;
	    }
}

HMineral::HMineral(GameObj* gob) : Hardware(gob) {
    SIZE = state.MINERAL_SIZE;
    CAN_MOVE = false;
    HW_TYPE = HW_MINERAL;
    INVINCIBLE = true;
    afterInit();
    c_debug << "mineral at (" << pos.x << ", " << pos.y << ")" << endl;
}

HWorker::HWorker(GameObj* gob) : Hardware(gob) {
    SIZE = state.WORKER_SIZE;
    HW_TYPE = HW_WORKER1;
    INVINCIBLE = true;
    afterInit();
}

AIObject::AIObject(GameObj* gob) {
    hardware = NULL;
    software = NULL;
    ignored = false;
    string name = gob->bp_name();
    owner = *gob->sod.owner;
    if (name == "sheep") {
	hardware = new HSheep(gob);
    } else if (name == "marine") {
	hardware = new HMarine(gob);
    } else if (name == "tank") {
	hardware = new HTank(gob);
    } else if (name == "controlCenter") {
	hardware = new HCC(gob);
    } else if (name == "mineral") {
	hardware = new HMineral(gob);
    } else if (name == "worker") {
	hardware = new HWorker(gob);
    } else {
	c_error << "Unknown unit type: " << name << endl;
	ERR("Unknown unit type");
    }
    hardware->aiob = this;
    if (owner == state.player_id) {
	software = new L3(hardware);
    } else {
	software = new dummyAI(hardware);
    }
}

void AIObject::update() {
    hardware->update();
    software->update();
}

void dummyAI::update() {
    TPoint p = hardware->pos;
    RCHK(NODEAT(p), MAX_NODES);
    state.used_nodes[NODEAT(p)] = hardware;
    for (int r=3; r<(hardware->CAN_MOVE ? 10 : 5); r+=3)
	for (int a=0; a<32; a++) {
	    int node = NODEAT(angle_r(p, a, r));
	    RCHK(node, MAX_NODES);
	    state.used_nodes[node] = hardware;
	}
}
