#ifndef Hardware_H
#define Hardware_H

#include "GameObj.H"

#include "TPoint.H"
#include "Common.H"
#include "L3.H"

#define HW_UNKNOWN	0
#define HW_SHEEP	1
#define HW_MARINE	2
#define HW_TANK		3
#define HW_CC		4
#define HW_MINERAL	5
#define HW_WORKER1	6
#define HW_WORKER2	7

using namespace std;

class AIObject;

class Hardware {
    protected:
	Hardware* detectAttacker(float e);
    public:
	int id;
	AIObject* aiob;			// XXX unused (?)
	int MAX_HP;
	int MAX_COOLDOWN;
	int SIZE;
	int RANGE;
	bool CAN_MOVE;
	bool CAN_SHOOT;
	bool INVINCIBLE;
	int HW_TYPE;			// XXX unused (?)
	bool dead;
	TPoint pos;			// unit's position
	int old_hp[HP_HISTORY_LENGTH];  // unit's hitpoints (n frames ago)
	int hp;				// unit's hitpoints (current)
	int cooldown;			// weapon's cooldown (current); 0 <=> ready to fire
	bool ally;			// true iff it's a player's unit
	Hardware* target;		// unit's direct target (null if unit is not attacking)
	GameObj* gob;
	Hardware(GameObj* gob);
	virtual ~Hardware() {};
	void afterInit();
	float distTo(Hardware* h);	// distance_to from gob
	bool isMoving();		// current speed > 0
	virtual void update();		// update unit's data from server (from gob)
};

class HSheep : public Hardware {	// a sheep
    public:
	HSheep(GameObj* gob);
	~HSheep() {};
};

class HMarine : public Hardware {	// a marine
    public:
	HMarine(GameObj* gob);
	~HMarine() {};
};

class HTank : public Hardware {		// a tank
    public:
	HTank(GameObj* gob);
	~HTank() {};
};

class HCC : public Hardware {		// a control center
    public:
	int x1;
	int y1;
	int x2;
	int y2;
	HCC(GameObj* gob);
	~HCC() {};
	void update();
};

class HMineral : public Hardware {	// a mineral
    public:
	HMineral(GameObj* gob);
	~HMineral() {};
};

class HWorker : public Hardware {	// a worker
    public:
	HWorker(GameObj* gob);
	~HWorker() {};
};

class AIObject {
    public:
	bool ignored;
	Hardware* hardware;		// phisical properties
	L3* software;			// unit's ai
	int owner;
	AIObject(GameObj* gob);
	~AIObject() {};
	void update();
};

class dummyAI : public L3 {		// ai for enemy and neutral unit's; do not follow orders
    public:
	dummyAI(Hardware* h) : L3(NULL) {hardware=h;};
	~dummyAI() {};
	void update();
};

#endif
