#include "Game.H"
#include "GameObj.H"

#include "Common.H"
#include "TPoint.H"
#include "AIState.H"
#include "AIGUI.H"
#include "Hardware.H"
#include "L1.H"

using namespace std;

L1::L1(Hardware* h) {
    if (h == NULL)
	return; // for dummy ai
    idle = true;
    hardware = h;
    low_target = NULL;
    low_stop();
    low_order = LOW_IDLE;
    load = 0;
}

bool L1::canAttack(AIObject* t) {
    if (t == NULL)
	return false;
    float d = dist(hardware->pos, t->hardware->pos) - hardware->SIZE - t->hardware->SIZE;
    return d < hardware->RANGE;
}

bool L1::canGather(AIObject* m) {
    if (m == NULL)
	return false;
    float d = hardware->distTo(m->hardware);
    return d < 3;
}

bool L1::canReturn(AIObject* cc) {
    if (cc == NULL)
	return false;
    float d = hardware->distTo(cc->hardware);
    return d < 4;
}

void L1::stopNow() {
    Vector<sint4> params;
    hardware->gob->set_action("stop", params);
}

void L1::moveNow(TPoint d) {
    Vector<sint4> params;
    params.push_back((int)(d.x));
    params.push_back((int)(d.y));
    int speed = (int)(*hardware->gob->sod.max_speed * (1 - (state.handicap / (float)100)));
    c_debug << "speed = " << speed << endl;
    params.push_back(speed);
    hardware->gob->set_action("move", params);
}

void L1::attackNow(AIObject* t) {
    Vector<sint4> params;
    params.push_back(state.gsm->get_game().get_cplayer_info().get_id(t->hardware->gob));
    hardware->gob->component("weapon")->set_action("attack", params);
}

void L1::gatherNow(AIObject* aiob) {
    load = hardware->gob->get_int("minerals");
    c_debug << "load = " << load << endl;
    if (load >= MAX_WORKER_LOAD) {
	hardware->HW_TYPE = HW_WORKER2;
    }
    Vector<sint4> params;
    params.push_back(state.gsm->get_game().get_cplayer_info().get_id(aiob->hardware->gob));
    hardware->gob->component("pickaxe")->set_action("mine", params);
}

void L1::returnNow(AIObject* aiob) {
    state.total_resources += load;
    load = 0;
    hardware->HW_TYPE = HW_WORKER1;
    Vector<sint4> params;
    params.push_back(state.gsm->get_game().get_cplayer_info().get_id(aiob->hardware->gob));
    hardware->gob->set_action("return_resources", params);
}

void L1::low_stop() {
    c_debug << hardware->id << ": low_stop" << endl;
    low_order = LOW_STOP;
    low_dest = nowhere;
    low_target = NULL;
    hardware->target = NULL;
}

void L1::low_move(TPoint d) {
    c_debug << hardware->id << ": low_move " << d.toString() << endl;
    low_order = LOW_MOVE;
    low_dest = d;
    low_target = NULL;
    hardware->target = NULL;
}

void L1::low_attack(AIObject* t) {
    c_debug << hardware->id << ": low_attack" << endl;
    low_order = LOW_ATTACK;
    low_dest = nowhere;
    low_target = t;
    hardware->target = t->hardware;
}

void L1::low_gather(AIObject* m) {
    c_debug << hardware->id << ": low_gather" << endl;
    if (m->hardware->HW_TYPE != HW_MINERAL)
	ERR("Wrong AIObject type");
    if (hardware->HW_TYPE != HW_WORKER1)
	ERR("Cannot gather");
    low_order = LOW_GATHER;
    low_dest = nowhere;
    low_target = m;
}

void L1::low_return(AIObject* cc) {
    c_debug << hardware->id << ": low_return" << endl;
    if (cc->hardware->HW_TYPE != HW_CC)
	ERR("Wrong AIObject type");
    if (hardware->HW_TYPE != HW_WORKER2)
	ERR("Cannot return");
    low_order = LOW_RETURN;
    low_dest = nowhere;
    low_target = cc;
}

void L1::low_commit() {
    c_debug << hardware->id << ": low_commit" << endl;
    if ((low_order == LOW_IDLE) || (low_order == LOW_STOP))
	stopNow();
    if (low_order == LOW_MOVE)
	moveNow(low_dest);
    if (low_order == LOW_ATTACK)
	attackNow(low_target);
    if (low_order == LOW_GATHER)
	gatherNow(low_target);
    if (low_order == LOW_RETURN)
	returnNow(low_target);
    low_order = LOW_IDLE;
}
