#include "Common.H"
#include "TPoint.H"
#include "AIState.H"
#include "Graph.H"
#include "Hardware.H"
#include "AIGUI.H"
#include "L1.H"
#include "L2.H"

using namespace std;

L2::L2(Hardware* h) : L1(h) {
    ready = true;
    init = false;
    prev_node = -1;
    curr_node = -1;
    next_node = -1;
    GATHER_RANGE = state.params.elen1 * 1.5;
    mode = MODE_ATTACK;
    paused = false;
    ignore = false;
    time_stopped = 0;
    face_dir = -1;
    min_dst = -1;
}

void L2::sync() {
    next_node = state.map->nodeAt(hardware->pos);
    next_node = state.map->nearbyOk(next_node);
    ready = false;
    init = true;
}

void L2::update() {
    if ((curr_node >= 0) && (curr_node < state.map->size))
	if (state.used_nodes[curr_node] == NULL)
	    state.used_nodes[curr_node] = hardware;
    if ((next_node >= 0) && (next_node < state.map->size))
	if (state.used_nodes[next_node] == NULL)
	    state.used_nodes[next_node] = hardware;
    if (hardware->isMoving()) {
	time_stopped = 0;
    } else {
	time_stopped++;
    }
}

void L2::move(int node) {
    if (ignore || paused)
	return;
    next_node = node;
    ready = false;
}

TPoint L2::correctDir(TPoint t) {
    const int maxr = 100;
    TPoint f = hardware->pos;
    int dx = (int)f.x-maxr/2;
    int dy = (int)f.y-maxr/2;
    bool ok[maxr][maxr];
    for (int i=0; i<maxr; i++)
	for (int j=0; j<maxr; j++)
	    if (RC2D(i+dx, j+dy, mapW, mapH)) {
		ok[i][j] = false;
	    } else {
		ok[i][j] = state.accessible[i+dx][j+dy];
	    }
    FOREACH(state.allUnits, it)
	if (!(((it->hardware->HW_TYPE == HW_MINERAL) && (hardware->HW_TYPE == HW_WORKER1))
		|| ((it->hardware->HW_TYPE == HW_CC) && (hardware->HW_TYPE == HW_WORKER2)))
	    && (it->hardware != hardware)) {
	    TPoint p = it->hardware->pos;
	    if (RC2D(p.x-dx, p.y-dy, maxr, maxr))
		continue;
	    ok[(int)(p.x-dx)][(int)(p.y-dy)] = false;
	    int r = it->hardware->SIZE;
	    for (int a=0; a<32; a++) {
		TPoint n = angle_r(p-TPoint(dx, dy), a, r);
		if (RC2D(n.x, n.y, maxr, maxr))
		    continue;
		ok[(int)(n.x)][(int)(n.y)] = false;
		n = angle_r(p-TPoint(dx, dy), a, r+1);
		if (RC2D(n.x, n.y, maxr, maxr))
		    continue;
		ok[(int)(n.x)][(int)(n.y)] = false;
	    }
	}
    int a0 = (int)angle(f, t);
    int r = hardware->SIZE + 2;
	for (int a=0; a<32; a++) {
	    TPoint n = angle_r(f, a+a0, state.params.look_ahead) - TPoint(dx, dy);
	    bool ok1 = true;
	    for (int a1=0; a1<32; a1++) {
		TPoint n1 = angle_r(n, a1, r);
		if (!ok[(int)(n1.x)][(int)(n1.y)])
		    ok1 = false;
	    }
	    TPoint p1 = hardware->pos;
	    TPoint p2 = angle_r(p1, a+a0, 30);
	    if (ok1) {
		if (a > 0)
		    face_dir = a+a0;
		return n + TPoint(dx, dy);
	    }
	}
    return nowhere;
}

AIObject* L2::chooseTarget() {
    AIObject* best_target = NULL;
    FOREACH(state.enemyUnits, opp) {
	if (!canAttack(opp))
	    continue;
	if (!canSee(hardware->pos, opp->hardware->pos))
	    continue;
	if (best_target == NULL)
	    best_target = opp;
	if ((!best_target->hardware->CAN_SHOOT)
	    && (state.oppTanks->size * 2 > state.myTanks->size)
	    && (opp->hardware->CAN_SHOOT))
		best_target = opp;
	if (opp->hardware->hp < best_target->hardware->hp)
	    best_target = opp;
    }
    if (best_target != NULL)
	return best_target;
    FOREACH(state.enemyUnits, opp) {
	if (!canAttack(opp))
	    continue;
	if (best_target == NULL)
	    best_target = opp;
	if ((!best_target->hardware->CAN_SHOOT)
	    && (state.oppTanks->size * 2 > state.myTanks->size)
	    && (opp->hardware->CAN_SHOOT))
		best_target = opp;
	if (opp->hardware->hp < best_target->hardware->hp)
	    best_target = opp;
    }
    return best_target;
}

void L2::commit() {
    if (paused || (mode == MODE_STOP))
	return;
    if ((mode > MODE_MOVE) && (state.enemyUnits->size > 0)) {
	AIObject* opp = chooseTarget();
	if (mode == MODE_STOPANDATTACK)
	    if (canAttack(opp)) {
		low_attack(opp);
		low_commit();
		return;
	    }
	if (mode == MODE_ATTACK) {
	    if (canAttack(opp) && (hardware->cooldown < 1)) {
		low_attack(opp);
		low_commit();
		return;
	    }
	}
    }
    if (next_node >= 0) {
        TPoint destp = state.map->nodePos(next_node);
	if (dist(hardware->pos, destp) < 1) {
	    prev_node = curr_node;
	    curr_node = next_node;
	    ready = true;
	    init = false;
	} else {
	    if (init && (!hardware->isMoving()) && (rand() % 2))
		sync();
	    low_move(destp);
	}
    }
    // WORKERS AI
    if ((hardware->HW_TYPE == HW_WORKER1) && (state.minerals->size > 0)) {
	AIObject* m = state.minerals->nearest(hardware->pos);
	float d = hardware->distTo(m->hardware);
	if (d <= GATHER_RANGE) {
	    ready = ignore = true;
	    next_node = -1;
	    if (d >= 1) {
		TPoint p = correctDir(m->hardware->pos);
		if (p != nowhere)
		    low_move(p);
	    } else {
		low_gather(m);
		if (load >= MAX_WORKER_LOAD-1) {
		    c_debug << hardware->id << " load" << endl;
		    mode = MODE_SYNC;
		}
	    }
	}
    } else if (mode == MODE_SYNC) {
	float dst = MAXINT;
	FOREACH(state.allUnits, it)
	    if (it->hardware->id != hardware->id)
		dst = MIN(dst, hardware->distTo(it->hardware));
	if (dst < hardware->SIZE - 1) {
	    c_debug << hardware->id << " sync stage 1" << endl;
	    low_move(angle_r(hardware->pos, state.game_time % 32, 5));
	} else {
	    if (dist(ROUND(hardware->pos), hardware->pos) > 2) {
		c_debug << hardware->id << " sync stage 2" << endl;
		low_move(ROUND(hardware->pos));
	    } else {
		c_debug << hardware->id << " sync stage 3" << endl;
		mode = MODE_MOVE;
	    }
	}
    }
    if (hardware->HW_TYPE == HW_WORKER2) {
	AIObject* cc = state.myControlCenters->any();
	float d = hardware->distTo(cc->hardware);
	if (d <= GATHER_RANGE) {
	    ready = ignore = true;
	    if (d >= 3) {
		TPoint p = correctDir(cc->hardware->pos);
		if (p != nowhere)
		    low_move(p);
	    } else {
		low_return(cc);
		mode = MODE_SYNC;
		c_debug << hardware->id << " return" << endl;
	    }
	}
    }
    low_commit();
}
