#include "TPoint.H"
#include "AIState.H"
#include "Hardware.H"
#include "UnitGroup.H"

using namespace std;

UnitGroup::UnitGroup() {
    size = 0;
    max_life = 0;
}

void UnitGroup::addUnit(AIObject* aiob) {
    c_debug << "adding unit: " << aiob->hardware->gob->bp_name();
    c_debug << " for player " << *aiob->hardware->gob->sod.owner << endl;
    v.push_back(aiob);
    size = v.size();
    RCHK(size, MAX_UNITS);
}

void UnitGroup::removeUnit(GameObj* gob) {
    if (gob == NULL)
	ERR("UnitGroup::removeUnit(NULL)");
    FORALL(v, it) {
	if ((*it)->hardware->gob == gob) {
	    v.erase(it);
	    size = v.size();
	    return;
	}
    }
    c_error << "(UnitGroup::removeUnit) error: unit not found!" << endl;
}

void UnitGroup::addGroup(UnitGroup* ug) {
    FOREACH(ug, it)
	addUnit(it);
}

int UnitGroup::totalLife(int when) {
    int l = 0;
    FOREACH(this, it) {
	if (!it->hardware->INVINCIBLE)
	    l += it->hardware->old_hp[when];
    }
    return l;
}

int UnitGroup::maxLife() {
    int l = 0;
    FOREACH(this, it) {
	if (!it->hardware->INVINCIBLE)
	    l += it->hardware->MAX_HP;
    }
    max_life = MAX(l, max_life);
    return max_life;
}

AIObject* UnitGroup::nearest(TPoint p, int n) {
    ASSERT(size > 0, "Unit group is empty");
    bool used[MAX_UNITS];
    int ic = 0;
    for (int i=0; i<size; i++)
	if (used[i] = v[i]->ignored)
	    ic++;
    n = MIN(n, size-ic-1);
    int c = -2;
    for (int i=0; i<=n; i++) {
	float mind = MAXINT;
	c = -1;
	for (int u=0; u<size; u++)
	    if (!used[u]) {
		float d = dist(p, v[u]->hardware->pos);
		if (d < mind) {
		    mind = d;
		    c = u;
		}
	    }
	RCHK(c, MAX_UNITS);
	used[c] = true;
    }
    ASSERT(c >= 0, "");
    return v[c];
}

float UnitGroup::nearest(int node, int n) {
    TPoint pos = NODEPOS(node);
    AIObject* aiob = nearest(pos, n);
    return dist(pos, aiob->hardware->pos);
}

float UnitGroup::distTo(UnitGroup* g) {
    float d = 0;
    FOREACH(this, it)
	d += g->nearest(it->software->curr_node);
    return d / size;
}

AIObject* UnitGroup::any() {
    return v[rand() % size];
}

TPoint UnitGroup::middle() {
    TPoint mid = TPoint(0, 0);
    FOREACH(this, it)
	mid = mid + it->hardware->pos;
    mid = mid / size;
    return mid;
}

int UnitGroup::busy() {
    int n = 0;
    FOREACH(this, it)
	if (!it->software->idle)
	    n++;
    return n;
}

int UnitGroup::outOfSync() {
    int n = 0;
    FOREACH(this, it)
	if ((!it->software->ready) && (it->software->time_stopped < 20))
	    n++;
    return n;
}

void UnitGroup::refresh() {
    FORALL(v, it) {
	if (((*it)->hardware->gob->is_dead())
	    || (!(*it)->hardware->gob->sod.in_game)) {
		c_info << "removing dead unit " << endl;
		NOTNULL((*it)->hardware);
		(*it)->hardware->dead = true;
		removeUnit((*it)->hardware->gob);
	}
    }
}
