// $Id: Net_DB.H,v 1.1 2005/08/23 21:24:53 hakman Exp $

// (c) Michael Buro 2003, mburo@cs.ualberta.ca, licensed under the GPL

/**
 * \class uses gdb to store/retrieve data
 * \author Sami Wagiaalla
 * \brief
 */

#ifndef NET_DB_H
#define NET_DB_H

#include "Global.H"
#include <gdbm.h>

#include "Net_DBTypes.H"

class Net_DB {

public: 

  Net_DB(const std::string& file);
  virtual ~Net_DB();

  /**
     adds the given player to the data base
     overwrites it if it already exists
     @return -1 of operation fails
  */
  sint4 put(DB_Type& p);

  /**
     gets the given player from the data base
     and puts the information in p
     @return -1 if error occures or player is not found
  */
  sint4 get(DB_Type& p);

  /**
     deletes the given player
     puts the deleted information in p
     @returns -1 of error occures or entry does not exist
  */
  sint4 rem(DB_Type& p);

  /**
     prints both data bases
  */
  void print();

  /**
     @return whether the give entry exists or not
  */
  bool exists(DB_Type& p);

  void close();
  
protected:
    
  /**
     copies the data in the given std::string to the given datum
     @return -1 if memory cannot be allocated
  */
  sint4 getData(const std::string& datas, datum& data);

  GDBM_FILE servicedb;
  std::string db_file;
};

#endif

