#include"linalg.h"

mat4 mat4::operator *(const mat4 &c) const
{
  mat4 r;
  for(int i=0;i<4;i++)
    for(int j=0;j<4;j++)
    {
      r[i][j]=0;
      for(int k=0;k<4;k++)
	r[i][j]+=A[i][k]*c.A[k][j];
    }
  return r;
};

vec4 mat4::operator *(const vec4 &v) const
{
  vec4 r;
  for(int i=0;i<4;i++)
  {
    r[i]=0;
    for(int j=0;j<4;j++)
      r[i]+=A[i][j]*v[j];
  }
  return r;
};

vec3 mat4::operator *(const vec3 &v) const
{
  vec3 r;
  for(int i=0;i<3;i++)
  {
    r[i]=0;
    for(int j=0;j<3;j++)
      r[i]+=A[i][j]*v[j];
    r[i]+=A[i][4];
  }
  return r;
};

mat4 mat4::trans()
{
  mat4 r;
  r.A[0][0]=A[0][0];
  r.A[0][1]=A[1][0];
  r.A[0][2]=A[2][0];
  r.A[0][3]=A[3][0];

  r.A[1][0]=A[0][1];
  r.A[1][1]=A[1][1];
  r.A[1][2]=A[2][1];
  r.A[1][3]=A[3][1];

  r.A[2][0]=A[0][2];
  r.A[2][1]=A[1][2];
  r.A[2][2]=A[2][2];
  r.A[2][3]=A[3][2];

  r.A[3][0]=A[0][3];
  r.A[3][1]=A[1][3];
  r.A[3][2]=A[2][3];
  r.A[3][3]=A[3][3];

  return r;
};


mat4 mat4::inv()
{
  mat4 r;
  r.A[0][0]=A[0][0];
  r.A[0][1]=A[1][0];
  r.A[0][2]=A[2][0];
  r.A[0][3]=-A[0][3];

  r.A[1][0]=A[0][1];
  r.A[1][1]=A[1][1];
  r.A[1][2]=A[2][1];
  r.A[1][3]=-A[1][3];

  r.A[2][0]=A[0][2];
  r.A[2][1]=A[1][2];
  r.A[2][2]=A[2][2];
  r.A[2][3]=-A[2][3];

  r.A[3][0]=0;
  r.A[3][1]=0;
  r.A[3][2]=0;
  r.A[3][3]=1;

  return r;
};

void mat4::identity()
{
  mat4 r;
  A[0][0]=1; A[0][1]=0; A[0][2]=0; A[0][3]=0;
  A[1][0]=0; A[1][1]=1; A[1][2]=0; A[1][3]=0;
  A[2][0]=0; A[2][1]=0; A[2][2]=1; A[2][3]=0;
  A[3][0]=0; A[3][1]=0; A[3][2]=0; A[3][3]=1;
};
