#ifndef DECOMP_NODE_ITEM_H
#define DECOMP_NODE_ITEM_H

// $Id: DecompUtil.H,v 1.2 2005/05/28 19:14:23 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include"Decomp.H"

// for copying the decomposition

class Decomp_Node_Item{
    
  typedef Vector<int> IList;

public:
  Decomp::Node node;
    
  int search_node;
  int lt, ge;
    
  IList neighbour_list[4];
    
  Decomp_Node_Item() {}
  ~Decomp_Node_Item() {}

  void set_node(Decomp::Node *n) {

    FORS (i, 4) neighbour_list[i].clear();

    if (n->type==Decomp::Node::R_NODE) {
      node.type = Decomp::Node::R_NODE;
      node.visited = n->visited;
      node.data = n->data;
      node.r = n->r;

      node.id = n->id;
      node.sub_id = n->sub_id;
      node.id0 = n->id0;
      node.sub_id0 = n->sub_id0;
		
    } else {
      node.type = n->type;
      node.visited = n->visited;
      node.v = n->v;
		
    }
  }
};
#endif

#ifndef DECOMPUTIL_H
#define DECOMPUTIL_H

#define buffer_size 5000

class DecompUtil
{
  Decomp* decomposition;

  int size;

  int number_of_items;
    
  Decomp_Node_Item* buffer;
    
public:
    
  DecompUtil():decomposition(NULL) {
    buffer = new Decomp_Node_Item[buffer_size];
    size = buffer_size;
  }

  ~DecompUtil() {
    delete buffer;
  }

  void attatch_decomp(Decomp *dec) {
    decomposition = dec;
  }

  void detach_decomp() {
    decomposition = NULL;
  }
  
  int get_node_id(Decomp::Node *node) {
    if (node==NULL) return -1;

    int index = 0;
    bool found = false;
    FORALL (decomposition->nodes, it) {
      if ((*it)==node) {
	found = true;
	break;
      }
      index++;
    }
    if (found) {
      return index;
    } else {
      return -1;
    }
  }

  Decomp* return_copy() {

    if (decomposition == NULL) return NULL;

    number_of_items = decomposition->nodes.size();
	  
    Decomp* new_decomp = new Decomp();
	  
    if ((int)(decomposition->nodes.size()) > size) {
      delete buffer;

      size = decomposition->nodes.size();
      buffer = new Decomp_Node_Item[size];
    }

    new_decomp->set_r0(decomposition->get_r0());

    int current = 0;
    FORALL(decomposition->nodes, it) {

      buffer[current].set_node(*it);
      if ((*it)->type != Decomp::Node::R_NODE) {
	buffer[current].lt = get_node_id((*it)->lt);
	buffer[current].ge = get_node_id((*it)->ge);
		    
      } else {
	FORS (i, 4) {
	  FORALL ((*it)->nl[0], j) {
	    buffer[current].neighbour_list[i].push_back(get_node_id(*j));
	  }
	}
      }
      current++;
		    
    }

    for (int i=0; i < number_of_items; i++) { 

      if (buffer[current].node.type == Decomp::Node::R_NODE) {

	// fill in the neighbour list
	FORS (j, 4) {
	  FORALL (buffer[current].neighbour_list[j],it) {
	    int index = *it;
	    buffer[current].node.nl[j].push_back(&(buffer[index].node));
	  }
	}
      } else {
	int lt = buffer[current].lt;
	int ge = buffer[current].ge;
	if (lt == -1) {
	  buffer[current].node.lt = NULL;
	} else {
	  buffer[current].node.lt = &(buffer[lt].node);
	}

	if (ge == -1) {
	  buffer[current].node.ge = NULL;
	} else {
	  buffer[current].node.ge = &(buffer[ge].node);
	}
      }

      new_decomp->nodes.push_back(&buffer[current].node);
    }

    //root node
    int root = get_node_id(decomposition->get_root());
	  
    new_decomp->set_root(&(buffer[root].node));

    return new_decomp;
  }

};
#endif
