#ifndef GfxTerrainTile_H
#define GfxTerrainTile_H

// $Id: GfxTerrainTile.H,v 1.37 2007/03/21 17:16:02 orts_mburo Exp $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxGlobal.H"
#include "GfxObject.H"
#include "Image.H"
#include "Tile.H"
#include "Shader.H"

class GfxTerrain;
class WaterTile;

//===================================================================

struct TileTex
{
  GLuint id;
  real4 (*coord)[4][2]; // [4][4][2]

  TileTex() { id = 0; coord = 0; }
};

//===================================================================

class GfxTerrainTile : public GfxObject
{
public:

  boost::array<real4,4> h; // corner heights (SW, SE, NW, NE)
  boost::array<sint1,4> flattened; ///< If the corner is ignoring noise.  Indices match h[4].


  /// Affects how the tile geometry is created and how texture is applied.
  /// Water tiles have trenches, ground and cliff tiles do not.
  enum GeometryType { UNKNOWN, WATER, GROUND, CLIFF };

  GfxTerrainTile();
  virtual ~GfxTerrainTile() {}

  void init(sint4 index, Game *game, GfxTerrain *t);
  void initblack(sint4 index, GfxTerrain *t);
  void initverts();
  bool is_initialized() const { return initialized; }

  // Announce changes to surrounding tiles.
  void NW_neighbor();
  void N_neighbor ();
  void NE_neighbor();
  void E_neighbor ();
  void SE_neighbor();
  void S_neighbor ();
  void SW_neighbor();
  void W_neighbor ();

  virtual void update(real4) {}

  virtual void draw();
  void draw_shader();
  
  void draw_flat();
  virtual void pick();
  virtual void draw_outline();

  static bool load_noise(const std::string &file,
			 const Vec3<real4> &scale1, const Vec3<real4> &scale2);
  static real4 get_noise_at(real4 x, real4 y);
  real4 height_at(real4 x, real4 y) const;

  void flatten(sint4 f);

  DrawOrder order() const { return O_TERRAIN; }

  void compute_tc();
  void setTexture(real4 row, real4 col=0,sint4 ncol=8,
                  float hflip=1, float vflip=1, bool transpose=0);

  bool has_water() { return (types[typeW] == WATER || types[typeE] == WATER); }
  WaterTile *new_WaterTile();

  static void add_type(sint4 id, GeometryType t);

  static bool outline;
  static ShaderProgram tile_shader;

protected:
  GfxTerrain *ter;
  sint4 index; ///< This tile's index in GfxTerrain.
  sint2 tx; // tile x
  sint2 ty; // tile y

  Vec3<real4> vert[4][4];
  Vec3<real4> norm[4][4];
  real4 texcoord[4][4][2];

  boost::array<real4,2> tt1; // texture tile 1, offset to grass (or other) texture
  real4 hf;     // texture tile 1 horizontal flip
  real4 vf;     // texture tile 1 vertical flip
  bool trans;   // texture tile 1 transpose 
  boost::array<real4,2> tt2; // texture tile 2, offset to cliff texture

  bool initialized;
  bool has_cliffs;
  Tile::Split split;
  Tile::Type typeW, typeE;


  inline void set_edge(sint4 a, sint4 b, real4 width, real4 depth);
  inline void set_edge(sint4 a, sint4 b, real4 width);

  GeometryType get_typeW() const { return types[typeW]; }
  GeometryType get_typeE() const { return types[typeE]; }
  // tE should be the same as tW for unsplit tiles, but don't take chances if split == NO_SPLIT
  GeometryType get_typeN() const { return (split == Tile::TB_SPLIT) ? types[typeE] : types[typeW]; }
  GeometryType get_typeS() const { return (split == Tile::BT_SPLIT) ? types[typeE] : types[typeW]; }

  
  static real4 sz;  // texture tile size
  static real4 eps; // small value (size of a pixel in the texture)

  static Image noise;
  static Vec3<real4> noise_scale1;
  static Vec3<real4> noise_scale2;

  static Vector<GeometryType> types;

  friend class GfxTerrain;
};

//===================================================================

#endif
