#ifndef Shader_H
#define Shader_H

// $Id: Shader.H,v 1.11 2007/04/13 15:03:44 orts_furtak Exp $

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "GfxGlobal.H"

//===================================================================

class Shader
{
public:
  GLuint shader;

  void clear();
  static char* read_source(const std::string &filename);
  void compile();
};

//-------------------------------------------------------------------

class VertexShader : public Shader
{
public:
  VertexShader() { shader = 0; }
  VertexShader(const std::string &filename) { load(filename); }

  bool load(const std::string &filename);

  ~VertexShader() {
    if (shader != 0) {
      glDeleteObjectARB(shader);
    }
  }
};

//-------------------------------------------------------------------

class FragmentShader : public Shader
{
public:
  FragmentShader() { shader = 0; }
  FragmentShader(const std::string &filename) { load(filename); }

  bool load(const std::string &filename);

  ~FragmentShader() {
    if (shader != 0) {
      glDeleteObjectARB(shader);
    }
  }
};

//-------------------------------------------------------------------

class ShaderProgram
{
public:
  ShaderProgram();
  ~ShaderProgram();

  GLuint id() const { return prog; }
  void fragment_shader(const std::string &filename);
  void vertex_shader(const std::string &filename);
  void reload();
  static void reload_all();

  void link();

  bool begin() {
    if (!glUseProgram || !prog) return false;
    glUseProgram(prog);
    return true;
  }

  void end() {
    if (glUseProgram) glUseProgram(0);
  }

protected:
  GLuint prog;
  FragmentShader *fs;
  VertexShader *vs;

  std::string fname, vname;

  static std::set<ShaderProgram*>& shaders();
};

REGISTER_TYPEOF(660, std::set<ShaderProgram*>::iterator);
//REGISTER_TYPEOF(661, std::set<ShaderProgram*>::const_iterator);

//===================================================================

#endif
