#ifndef PairHeap_H
#define PairHeap_H

// $Id: PairHeap.H,v 1.9 2006/04/25 03:08:38 orts_mburo Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Global.H"
#include "CycleCounter.H"

// fixme: better data structure => no map needed, store info with the edge

// heap of pairs (x,y)   x unique
//
// operations:
//
//  ins(x,y)
//  del(x)   !
//  get min(x,y)
//  get max(x,y)

template <class X, class Y> class PairHeap {

 private:

 protected:

 public:

  sint4 ins_n, del_n;
  uint4 ins_s, del_s;
  uint8 ticks;
  
  struct Pair {
    X x;
    Y y;
  };

  struct PairComp
  {
    bool operator()(const Pair &p1, const Pair &p2) const
    {
      if (p1.y < p2.y) return true;
      if (p2.y < p1.y) return false;
      return p1.x < p2.x;
    }
  };

  std::map<X,Y>  m;
  std::set<Pair,PairComp> s;

  PairHeap() {
    ins_n = del_n = 0;
    ins_s = del_s = 0;
    ticks = 0;
  }
  
  virtual ~PairHeap() {}

  uint4 size() const { return m.size(); }

  // return true <=> new (x,y)

  bool ins(const X &x, const Y &y) {

    ins_n++;
    //    CycleCounter start;

    // We can't use "FIND(m, it, x)" due to the "typeof" issues coupled with a template class
    typename std::map<X,Y>::iterator it = m.find(x);

    if (it != m.end()) {
      if (it->second == y) {
        //	CycleCounter end;
        //	ticks += end.count() - start.count();
	return false;
      }

      Pair p;
      p.x = x;
      p.y = it->second;
      s.erase(p);

      p.y = y;
      s.insert(p);
      it->second = y;

      assert(m.size() == s.size());
      
    } else {

      // new x

      m.insert(std::pair<X,Y>(x,y));
      Pair p;
      p.x = x;
      p.y = y;
      s.insert(p);
      
      assert(m.size() == s.size());
    }

    //    CycleCounter end;
    //    ticks += end.count() - start.count();
    return true;
  }

  // return true <=> deleted
  
  bool del(X &x) {
    del_n++;
    //    CycleCounter start;

    // We can't use "FIND(m, it, x)" due to the "typeof" issues coupled with a template class
    typename std::map<X,Y>::iterator it = m.find(x);

    if (it == m.end()) {
      //      CycleCounter end;
      //      ticks += end.count() - start.count();
      return false;
    }
    Pair p;
    p.x = x;
    p.y = it->second;
    m.erase(it);
    s.erase(p);
    assert(m.size() == s.size());
    //    CycleCounter end;
    //    ticks += end.count() - start.count();
    return true;
  }

  // return true <=> heap not empty
  bool min(X &x, Y &y) const {
    if (s.empty()) return false;
    typename std::set<Pair,PairComp>::const_iterator it = s.begin();
    x = it->x;
    y = it->y;
    return true;
  }

  // return true <=> heap not empty
  bool max(X &x, Y &y) const {
    if (s.empty()) return false;
    typename std::set<Pair,PairComp>::const_reverse_iterator it = s.rbegin();
    x = it->x;
    y = it->y;
    return true;
  }
};
  
#endif
