#ifndef BSC_OBJ_H
#define BSC_OBJ_H

#define OGRE_RELEASE_ASSERT 0

#include "GameObj.H"
#include "OgreGfxAnimationSet.H"
#include <string.h>
#include <Ogre.h>
#include <deque>
#include <iostream>

/**==================================================================================================
A base class for OgreObjects which should hold information that all objects will hold.  This should eventually
be an abstract class to OgreMD3Model once that is working 100% which atm is not the case.
==================================================================================================*/
class BasicObject{
public:
	/**Is the object selected in the GUI*/
        bool selectedObject;  

	/**Reference to the object in the gameworld see GameObj for what this contains*/
	GameObj *gameObj;

	/**The SceneManager or basically the 3d universe in ogre where everything is added*/
	Ogre::SceneManager* gdSceneMgr;

	/**The Name of the object in Ogre SceneManager(is unique)*/
	Ogre::String objName;					   //String name of the Object.
	
	/**The Entity*/
	Ogre::Entity *objEntity;					   //used to represent Object in Ogre Game Engine.(assumed created).
	/**The SceneNode which is used to rotate, translate, scale the object*/
	Ogre::SceneNode *objNode;
			
	/**The state of animation in the object*/		   
	Ogre::AnimationState *objAnimateState;		//state of animation of the object.
	/**Where teh object needs to walk too*/
	std::deque<Ogre::Vector3> walkPoints;	//point the Object Needs to Walk too.

	Ogre::Real objAnimateSpeed;					//Speed of the current animation.
	Ogre::Real objWalkSpeed;						//Speed the unit is walking.
	Ogre::Vector3 objWalkDestination;				//Where the Object is going too
	
	Ogre::Quaternion mOrientSrc;               // Initial orientation
	Ogre::Quaternion mOrientDest;              // Destination orientation
	Ogre::Vector3 rotSrc;
	Ogre::Vector3 direction;
	Ogre::Real mRotProgress;                   // How far we've interpolated
	Ogre::Real mRotFactor;                     // Interpolation step
	bool mRotating;						 // Is object rotating.

	std::deque<Ogre::Real> objWalkPoints;   //Deque of Points the Unit has to still go too.

	BasicObject(Ogre::String &objName, Ogre::SceneManager* gdSceneMgr, GameObj *gameObj, std::map<std::string, std::string> &bpName2MeshNameMap);  //ONLY constructor you need mesh name/fname mesh.
};

#endif

