#ifndef OgreGfxModuleMegaTile_h
#define OgreGfxModuleMegaTile_h

#define OGRE_RELEASE_ASSERT 0

#include "Ogre.h"
#include "OgreKeyEvent.h"
#include "OgreEventListeners.h"
#include "OgreStringConverter.h"
#include "OgreException.h"
#include <OgreStaticGeometry.h>
#include <vector>


class OgreGfxModuleMegaTile
{
public:

  /**
     Creates a new megatile
     \param strtr the megaTile's starting row
     \param strtc starting column
     \param r how many rows this megatile will have
     \param c how many columns
     \param tW how wide a tile is
     \param tH how high (z direction)
     \param revealed array of revealed tiles from the TerrainTileGrid
     \param visible  array of visible tiles from the TerrainTileGrid
     \param sceneMgr need the reference to the sceneManager
     \param cast_shadows should probably be false, terrain shouldn't cast shadows
   */
  OgreGfxModuleMegaTile(int strtr=0, int strtc=0,int r=4, int c=5, float tW=20.0f, float tH=20.0f, bool** revealed = NULL, bool** visible=NULL, Ogre::SceneManager* sceneMgr= NULL, bool cast_shadows = false);
  ~OgreGfxModuleMegaTile();
  
  /**
     Adds a new ogre entity to this megatile
   */
  void addGeometry(Ogre::Entity* newEnt, float x = 0.0f, float y = 0.0f,  float z = 0.0f, float xscale = 1.0f, float yscale = 1.0f, float zscale = 1.0f, Ogre::Quaternion orientation = Ogre::Quaternion::IDENTITY);
  
  /**
     Must rebuild in order for the new geometry to show up
   */
  void rebuild();
  
  int startr;//the tile row and column where this megatile starts
  int startc;
  int rows;//the number of rows and columns that make up this megatile
  int columns;
  float tileW;//the width and height of a single tile in this megatile
  float tileH;
  bool bRebuildNeeded;//indicates whether or not this megatile has been modified and is need of rebuilding.
  bool bCastShadows;
  
private:
  Ogre::StaticGeometry* megaTileGeometry;//the tile's static mesh geometry
  Ogre::SceneManager* mSceneMgr;//a pointer to the scene manager
};

#endif
