#ifndef OGRE_MD3H
#define OGRE_MD3H

#include "GfxGlobal.H"
#include "BasicObject.H"

#include <Ogre.h>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>

#define MD3_ID (((int)'I')|(((int)'D')<<8)|(((int)'P')<<16)|(((int)'3')<<24))

class MD3Vertex;
class MD3Header;
class MD3Surface;
class Poly;
class ModelTag;
class BBox;
class MD3Tag;

/*=================================================================================
OgreMD3Model - 
This class in future will extend BasicObject and be used inside ObjectManager as all 
objects of orts should be MD3.  Basicobject can handle (.mesh) skeleton animation atm
which is trivial.  But keyframe is difficult and although loader is working full,
implementation is not yet working.
===================================================================================*/
class OgreMD3Model{
public:
  sint4 load(const std::string &modelfile); //loads the Model from Modelfile MD3 
  void createMD3Object(Ogre::SceneManager* mgr, const std::string &modelName, const int scale=100); //converts and loads the model into ogre. 

private:
  void load_cfg(const std::string &fname);            //loads the .cfg file(called from load)
  bool load_header(FILE* &f, MD3Header &header);      //loads the header(called from load)
  bool load_tags(FILE* &f, const MD3Header &header);  //loads the tags(called from load)

  //hold info on what frames each animation is part of eg name2frames["idle"] will give pair 2,19 for mech.
  std::map<std::string, std::pair<sint2, sint2> > name2frames; 
  Vector<BBox> bbox;  //TODO:implement this w/ ogre.

  //Below variables are just temp create in load so that ogre objs can be created in createMD3 at which time they are deleted.
  uint4 nvert;
  uint4 npoly;
  Vector<std::string> tag_name;
  Vector<ModelTag*> tag; // each element is an array - ModelTag[nframe]
  ModelTag *all_tags;
  Vec3<real4> *vert;
  real4 *tvert;
  Vec3<real4> *vnorm;
  Poly *poly;

};

struct MD3Header{
  sint4 id;
  sint4 version;
  uint1 name[64];
  sint4 flags;
  sint4 numFrames;
  sint4 numTags;
  sint4 numSurfaces;
  sint4 numSkins;
  sint4 surfaceSize;
  sint4 offsetTags;     // tag start
  sint4 offsetSurfaces; // surface start / tag end
  sint4 offsetEnd;      // end of file ~ file size
  void print() const;
};

struct MD3Surface{
  sint4 id;
  uint1 name[64];
  sint4 flags;
  sint4 numFrames;
  sint4 numShaders;
  sint4 numVerts;
  sint4 numTriangles;
  sint4 offsetTriangles;
  sint4 offsetShaders;
  sint4 offsetST;
  sint4 offsetVerts;
  sint4 offsetEnd;
  void print() const;
};

struct MD3Tag{
  uint1 name[64];
  real4 origin[3];
  real4 axis[3][3];
};

struct MD3Vertex{
  sint2 x;
  sint2 y;
  sint2 z;
  sint2 normal;
};

struct Poly{
  sint2 vert[3];
};

struct ModelTag{
  Mat4<real4> axis;
};

struct BBox{
  Vec3<real4> min;
  Vec3<real4> max;
};

#endif
