// This is an ORTS file 
// (c) Michael Buro,
// (c) Sami Wagiaalla 
// licensed under the GPL

#include "Path.H"

UnitPath::UnitPath(GameObj *obj, const Vector<Point2i> &path_) : unit(obj), path(path_) {
  tries = -1;
  current_target = path.size()-1;
  dead = false;
  if (path.size() > 0) {
    goal = path[0];
  }
  spot = Point2i(-1, -1);
  wakeup = 0;
}


const double EPS = 1e-9;

Vector<Line> UnitPath::intersect(const Point2i &a, const Point2i &b)
{
  Vector<Line> lines;
  
  for (uint4 i = 1; i < path.size(); i++) {

    Point2i c = path[i-1];
    Point2i d = path[i];
    
    double m1 = (double)(c.y-d.y) / (double)(c.x-d.x + EPS);
    double m2 = (double)(a.y-b.y) / (double)(a.x-b.x + EPS);
    
    if ((int)m1 == (int)m2) continue; // fixme: equality for reals doesn't work
    
    double b1 = c.y - m1 * c.x;
    double b2 = a.y - m2 * a.x;        
    
    double X = (b2-b1) / (m1-m2);
    double Y = m1*X + b1;
    
    double t1x = (double)(X - a.x) / ((double)(b.x - a.x + EPS));
    double t1y = (double)(Y - a.y) / ((double)(b.y - a.y + EPS));
    double t2x = (double)(X - c.x) / ((double)(d.x - c.x + EPS));
    double t2y = (double)(Y - c.y) / ((double)(d.y - c.y + EPS));
    
    bool overlap1 = (t1x < 0 || t1x > 1);
    if (!overlap1) continue;
    
    bool overlap2 = (t1y < 0 || t1y > 1);
    if (!overlap2) continue;    
    
    bool overlap3 = (t2x < 0 || t2x > 1);
    if (!overlap3) continue;        

    bool overlap4 = (t2y < 0 || t2y > 1);
    if (!overlap4) continue;    

    lines.push_back(Line(path[i-1], path[i]));      
  }
  
  return lines;
}
