//-----------------------------------------------------------------------------
/** @file util.h
    General utilities.

    $Id: util.h,v 1.1 2006/12/31 04:09:42 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/pathfinding/ua_path_lib/src/util.h,v $
*/
//-----------------------------------------------------------------------------

#ifndef PATHFIND_UTIL_H
#define PATHFIND_UTIL_H

#include <ctime>
#include <iostream>
#include <string>
#include "error.h"
#include "math.h"

//-----------------------------------------------------------------------------

namespace PathFind
{
    /** Measure time with std::clock().
        Resolution and max range before overflow is platform dependent.
        Don't expect more than a msec resolution (even if CLOCKS_PER_SEC
        is higher) and a range of 1h.
    */
    class Clock
    {
    public:
        Clock()
        {
            reset();
        }

        void reset()
        {
            m_start = clock();
        }

        double get() const
        {
            // Automatically handles one overflow correctly no matter whether
            // clock_t is signed or unsigned
            clock_t diff = clock() - m_start;
            return static_cast<double>(diff) / CLOCKS_PER_SEC;
        }

    private:
        clock_t m_start;
    };

    /** Exception thrown by LineReader. */
    class ReadError
        : public Error
    {
    public:
        ReadError(int line, const string& info);
    };

    /** Wrapper class around std::istream for reading line by line.
        Allows keeping track of the line number for generating meaningful
        error messages.
    */
    class LineReader
    {
    public:
        static const int MAX_LINE = 2048;

        LineReader(std::istream& in);

        /** Creates a new error with a message and the current line number */
        Error createError(const string& message);

        int getLineNumber()
        {
            return m_lineNumber;
        }

        /** Read next line.
            @exception ReadError Reading failed.
        */
        std::string readLine();

    private:
        int m_lineNumber;

        std::istream& m_in;
    };

    /** Get a label useful for implementations of Search::getVisitedNodes.
        The char label is built with the iteration number and uses
        '0'-'9', 'a'-'z' if the iteration number is less than 37,
        '+' otherwise.
     */
    char getVisitedNodeLabel(int iteration);
}

//-----------------------------------------------------------------------------

#endif
