#ifndef GLOBALPATHFINDER_H
#define GLOBALPATHFINDER_H
// $Id: GlobalPathfinder.H,v 1.2 2006/09/13 00:12:29 nicholas Exp $

#include "Object.H"
#include "TerrainBase.H"

class GlobalPathfinder{
public:

  typedef TerrainBase::Loc Loc;
  typedef TerrainBase::Path Path;
  typedef TerrainBase::Segment Segment;
  typedef TerrainBase::ConsiderObjects ConsiderObjects;
  
  GlobalPathfinder() : tiles_x(0), tiles_y(0), tile_points(0), client_pID(0), neutral_pID(0), initiated(false){};

  virtual ~GlobalPathfinder(){};

  virtual void init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_,sint4 client_pID_, sint4 neutral_pID_){
    tiles_x = tiles_x_;
    tiles_y = tiles_y_;
    tile_points = tile_points_;
    client_pID = client_pID_;
    neutral_pID = neutral_pID_;
    initiated = true;
  };
  
  virtual void insert_object(const Object *obj) = 0;
  virtual void remove_object(const Object *obj)= 0;
  virtual void update_object(const Object *obj) = 0;
  virtual void insert_boundaries(const Vector<Segment> &s) = 0;
  
   /** Perform pathfinding for a given Task, return true iff path found, and the path itself in 'output'.
        If found then output is guaranteed to include at least one waypoint for the goal. 
        Waypoints are listed in order, where the goal is at locs[locs.size()-1]. */
  virtual bool find_path(const TerrainBase::Task& task, Path& output,  ConsiderObjects consider) = 0;
    
protected:
  sint4 tiles_x, tiles_y, tile_points, client_pID, neutral_pID;
  bool initiated;

};

#endif
