/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Client.H"
#include "System.H"

Client::~Client() { close(); }

bool Client::connect(const char *host_, sint4 port_)
{
  host = host_;
  port = port_;
  sb.set_socket(System::sock_connect(host_, static_cast<uint4>(port), false));
  if (sb.get_socket() <= 0) return true;
  return false;
}


void Client::close()
{
  REM("closing connection");

  if (!is_closed()) {
    System::sock_shutdown(sb.get_socket(), 2, false);
    System::sock_close(sb.get_socket(), false);
    sb.set_socket(SocketBuffer::NO_SOCKET);

  } else REM("!!! already closed");
}


bool Client::is_closed() const
{
  return sb.get_socket() == SocketBuffer::NO_SOCKET;
}



sint4 Client::write_flush() {
  return sb.blocking_flush();
}

sint4 Client::write_msg(const string &msg, bool blocking) {
  return sb.write_msg(msg, blocking);
}

sint4 Client::read_msg(string &msg) {
  return sb.read_msg(msg);
}

sint4 Client::write_line(const string &s, bool blocking) {
  string t = s;
  t += "\n";
  return sb.write(t.c_str(), t.size(), blocking);
}

sint4 Client::read_char(char &c, bool blocking) {
  return sb.read(&c, 1, blocking);
}

sint4 Client::read_msg2(string &msg, bool ready) {
  return sb.read_msg2(msg, ready);
}

