/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Global.H"
#include <map>

// fixme: parameterless options missing

class Opt {

public:

  bool orig;
  string remark;
  
  // true <=> error
  virtual bool from_string(const std::string &s) = 0;

  Opt() { orig = true; }
  virtual ~Opt() {}
};


template <class T> class OptT : public Opt
{
public:
  
  T value;

  OptT(T v) { value = v; }

  bool from_string(const std::string &s);
};


class Options {

protected:

  typedef map<std::string,Opt*> OMap;
  typedef OMap::iterator OMapIt;
  typedef OMap::const_iterator OMapCIt;  
  OMap omap;

public:

  template <class T>
  void put(const char *skey, T v, const char *remark=0)
  {
    string key(skey);
    OMapIt oit = omap.find(key);

    if (oit == omap.end()) {

      if (!remark) ERR("remark missing");
      
      // new entry
      OptT<T> *opt = new OptT<T>(v);
      opt->remark = string(remark);
      omap[key] = opt;
      return;
    }

    if (remark) ERR("new remark?");
    
    // entry found, check type

    OptT<T> *opt = dynamic_cast<OptT<T>*>(oit->second);

    if (!opt) ERR("illegal type");

    opt->orig = opt->value == v;
    opt->value = v;
  }

  // ERR iff illegal
  template <class T> 
  bool get(const std::string &key, T &v) const
  {
    OMapCIt oit = omap.find(key);

    if (oit == omap.end()) {
      // not found
      cerr << "option not found: " << key << endl;
      ERR("");
      return true;
    }      
    
    OptT<T> *opt = dynamic_cast<OptT<T>*>(oit->second);
    if (!opt) {
      // illegal type
      cerr << "illegal value type: " << key << endl;
      ERR("");      
      return true;
    }

    v = opt->value;
    return false;
  }

  bool process(sint4 argc, char **argv, ostream &oserr, const char *remark=0);
  
  virtual ~Options() {

    FORALL (OMap, omap, i) {
      if (i->second) { delete i->second; }
    }

  }
  
};

