/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef PairHeap_H
#define PairHeap_H

#include "Global.H"
#include <map>
#include <set>
#include "CycleCounter.H"

// fixme: better data structure => no map needed, store info with the edge

// heap of pairs (x,y)   x unique
//
// operations:
//
//  ins(x,y)
//  del(x)   !
//  get min(x,y)
//  get max(x,y)

template <class X, class Y> class PairHeap {

 private:

 protected:

 public:

  sint4 ins_n, del_n;
  uint4 ins_s, del_s;
  uint8 ticks;
  
  struct Pair {
    X x;
    Y y;
  };

  struct PairComp
  {
    bool operator()(const Pair &p1, const Pair &p2) const
    {
      if (p1.y < p2.y) return true;
      if (p2.y < p1.y) return false;
      return p1.x < p2.x;
    }
  };

  map<X,Y>  m;
  set<Pair,PairComp> s;

  PairHeap() {
    ins_n = del_n = 0;
    ins_s = del_s = 0;
    ticks = 0;
  }
  
  virtual ~PairHeap() {}

  uint4 size() const { return m.size(); }

  // return true <=> new (x,y)

  bool ins(const X &x, const Y &y) {

    ins_n++;
    CycleCounter start;

    typename map<X,Y>::iterator it = m.find(x);
    if (it != m.end()) {
      if (it->second == y) {
	CycleCounter end;
	ticks += end.count() - start.count();
	return false;
      }

      Pair p;
      p.x = x;
      p.y = it->second;
      s.erase(p);

      p.y = y;
      s.insert(p);
      it->second = y;

      assert(m.size() == s.size());
      
    } else {

      // new x

      m.insert(pair<X,Y>(x,y));
      Pair p;
      p.x = x;
      p.y = y;
      s.insert(p);
      
      assert(m.size() == s.size());
    }

    CycleCounter end;
    ticks += end.count() - start.count();
    return true;
  }

  // return true <=> deleted
  
  bool del(X &x) {
    ins_n++;
    CycleCounter start;

    typename map<X,Y>::iterator it = m.find(x);
    if (it == m.end()) {
      CycleCounter end;
      ticks += end.count() - start.count();
      return false;
    }
    Pair p;
    p.x = x;
    p.y = it->second;
    m.erase(it);
    s.erase(p);
    assert(m.size() == s.size());
    CycleCounter end;
    ticks += end.count() - start.count();
    return true;
  }

  // return true <=> heap not empty
  bool min(X &x, Y &y) {
    if (s.empty()) return false;
    typename set<Pair>::iterator it = s.begin();
    x = it->x;
    y = it->y;
    return true;
  }

  // return true <=> heap not empty
  bool max(X &x, Y &y) {
    if (s.empty()) return false;
    typename set<Pair>::reverse_iterator it = s.rbegin();
    x = it->x;
    y = it->y;
    return true;
  }
};
  
#endif

