// $Id: AI_Common.C,v 1.1.1.1 2004/10/11 09:09:20 orts_mburo Exp $

// This is an ORTS file (c) Michael Chung, licensed under the GPL

#include "Global.H"

#include "AI_Common.H"

#include "Game.H"
#include "GameObj.H"


//----------------------------------------------------------------------
// Utility functions
//----------------------------------------------------------------------

sint4 get_obj_type(const GameObj &gob)
{
  return gob.get_int("obj_type");
}

/** get position of an object (as a point) */
Point get_position(const GameObj &gob)
{
  return Point(gob.get_int("x"), gob.get_int("y"));
}

/** calculate distance between two points */
double calc_distance(const Point &a, const Point &b)
{
  double dx = b.x - a.x;
  double dy = b.y - a.y;
  
  return sqrt(dx*dx + dy*dy);
}

// alternative to Game::distance function
// doesn't exactly match the Game::distance function, but that doesn't
// seem to be working for me at the moment... (returns error?)
// the difference seems to be slight, probably just the unit radius
double calc_distance(const GameObj &a, const GameObj &b)
{
  if (TYPE_GLOBAL == get_obj_type(a)) ERR("!");
  if (TYPE_GLOBAL == get_obj_type(b)) ERR("!");
   
  Point ptA = get_position(a);
  Point ptB = get_position(b);

  return calc_distance(ptA, ptB);  
}

bool is_unit(const GameObj &gob)
{
  return (TYPE_UNIT == get_obj_type(gob));
}

bool has_flag(const GameObj &gob)
{
  return 0 != gob.get_int("has_flag");
}

/** returns board-position of an object */
Point get_board_position(const GameObj &gob)
{
  if (TYPE_GLOBAL == get_obj_type(gob)) ERR("!");
    
  return conv_to_board_point(get_position(gob));
}

bool is_equal(const Point &p1, const Point &p2)
{
  return (p1.x == p2.x) && (p1.y == p2.y);
}

/** returns if a unit is in motion */
bool is_moving(const GameObj &gob)
{
  if (!is_unit(gob)) {
    // all non-units assumed to be stationary
    return false;
  }
  
  return gob.get_int("speed");
}

/** return the hitpoints of a unit */
sint4 get_hp(const GameObj &obj)
{
  return obj.get_int("hp");
}

/** convert from a env point to a board point */
Point conv_to_board_point(const Point &pt)
{
  return Point(pt.x / AREA_SIZE, pt.y / AREA_SIZE);
}

/** convert from board point to env point */
Point conv_to_env_point(const Point &pt)
{
  return Point(
    (pt.x * AREA_SIZE) + (AREA_SIZE / 2), 
    (pt.y * AREA_SIZE) + (AREA_SIZE / 2));
}

Point snap_point_to_grid(const Point &pt)
{
  return Point(16 * (pt.x / 16) + 8, 16 * (pt.y / 16) + 8);
}

sint4 get_obj_id(const GameObj &gob, Game* game)
{
  return game->get_cplayer_info().get_id(const_cast<GameObj*>(&gob));
}

// Local Variables:
// c-file-style:"gnu"
// c-basic-offset:2
// indent-tabs-mode:nil
// tab-width:4
// End:
